import torch
from einops import rearrange
from torch import nn

from zeta.nn.attention.local_attention import LocalAttention
from zeta.utils.main import default, exists, l2norm


class LocalMHA(nn.Module):
    def __init__(
        self,
        *,
        dim,
        window_size,
        dim_head=64,
        heads=8,
        dropout=0.0,
        causal=False,
        prenorm=False,
        qk_rmsnorm=False,
        qk_scale=8,
        use_xpos=False,
        xpos_scale_base=None,
        exact_windowsize=None,
        **kwargs,
    ):
        super().__init__()
        inner_dim = dim_head * heads
        self.norm = nn.LayerNorm(dim) if prenorm else None

        self.heads = heads
        self.to_qkv = nn.Linear(dim, inner_dim * 3, bias=False)
        self.qk_rmsnorm = qk_rmsnorm

        if qk_rmsnorm:
            self.q_scale = nn.Parameter(torch.ones(dim_head))
            self.k_scale = nn.Parameter(torch.ones(dim_head))

        self.attn_fn = LocalAttention(
            dim=dim_head,
            window_size=window_size,
            causal=causal,
            autopad=True,
            scale=(qk_scale if qk_rmsnorm else None),
            exact_windowsize=default(exact_windowsize, True),
            use_xpos=use_xpos,
            xpos_scale_base=xpos_scale_base,
            **kwargs,
        )

        self.to_out = nn.Linear(inner_dim, dim, bias=False)

    def forward(self, x, mask=None, attn_bias=None):
        if exists(self.norm):
            x = self.norm(x)

        q, k, v = self.to_qkv(x).chunk(3, dim=-1)
        q, k, v = map(
            lambda t: rearrange(t, "b n (h d) -> b h n d", h=self.heads),
            (q, k, v),
        )

        if self.qk_rmsnorm:
            q, k = map(l2norm, (q, k))
            q = q * self.q_scale
            k = k * self.k_scale

        out = self.attn_fn(q, k, v, mask=mask, attn_bias=attn_bias)
        out = rearrange(out, "b h n d -> b n (h d)")
        return self.to_out(out)



import torch
from torch import nn, Tensor
from zeta.nn.modules.chan_layer_norm import ChanLayerNorm
from einops import rearrange


class ScalableImgSelfAttention(nn.Module):
    """
    ScalableImgSelfAttention module applies self-attention mechanism to image data.

    Args:
        dim (int): The input dimension of the image.
        heads (int, optional): The number of attention heads. Defaults to 8.
        dim_key (int, optional): The dimension of the key vectors. Defaults to 32.
        dim_value (int, optional): The dimension of the value vectors. Defaults to 32.
        dropout (float, optional): The dropout rate. Defaults to 0.0.
        reduction_factor (int, optional): The reduction factor for downscaling the image. Defaults to 1.

    Attributes:
        dim (int): The input dimension of the image.
        heads (int): The number of attention heads.
        dim_key (int): The dimension of the key vectors.
        dim_value (int): The dimension of the value vectors.
        reduction_factor (int): The reduction factor for downscaling the image.
        scale (float): The scaling factor for the key vectors.
        attend (nn.Softmax): The softmax function for attention calculation.
        dropout (nn.Dropout): The dropout layer.
        norm (ChanLayerNorm): The channel-wise layer normalization.
        to_q (nn.Conv2d): The convolutional layer for query projection.
        to_k (nn.Conv2d): The convolutional layer for key projection.
        to_v (nn.Conv2d): The convolutional layer for value projection.
        to_out (nn.Sequential): The sequential layer for output projection.

    """

    def __init__(
        self,
        dim: int,
        heads: int = 8,
        dim_key: int = 32,
        dim_value: int = 32,
        dropout: float = 0.0,
        reduction_factor: int = 1,
        *args,
        **kwargs,
    ):
        super().__init__()
        self.dim = dim
        self.heads = heads
        self.dim_key = dim_key
        self.dim_value = dim_value
        self.reduction_factor = reduction_factor

        self.scale = dim_key**-0.5
        self.attend = nn.Softmax(dim=-1)
        self.dropout = nn.Dropout(dropout)
        self.norm = ChanLayerNorm(dim)

        # Projections
        self.to_q = nn.Conv2d(dim, dim_key * heads, 1, bias=False)
        self.to_k = nn.Conv2d(
            dim,
            dim_key * heads,
            reduction_factor,
            stride=reduction_factor,
            bias=False,
        )
        self.to_v = nn.Conv2d(
            dim,
            dim_value * heads,
            reduction_factor,
            stride=reduction_factor,
            bias=False,
        )

        self.to_out = nn.Sequential(
            nn.Conv2d(dim_value * heads, dim, 1), nn.Dropout(dropout)
        )

    def forward(self, x: Tensor) -> Tensor:
        """
        Forward pass of the ScalableImgSelfAttention module.

        Args:
            x (Tensor): The input tensor of shape (batch_size, channels, height, width).

        Returns:
            Tensor: The output tensor of shape (batch_size, channels, height, width).

        """
        h, w, h = *x.shape[-2:], self.heads

        x = self.norm(x)

        q, k, v = self.to_q(x), self.to_k(x), self.to_v(x)

        # Split out heads
        q, k, v = map(
            lambda t: rearrange(t, "b (h d) ... -> b h (...) d", h=h),
            (
                q,
                k,
            ),
        )

        # Similarity
        dots = torch.matmul(q, k.transpose(-1, -2)) * self.scale

        # Attention
        attn = self.attend(dots)
        attn = self.dropout(attn)

        # Aggregate values
        out = torch.matmul(attn, v)

        # Merge back heads
        out = rearrange(
            out,
            "b h (x y) d -> b (h d) x y",
            x=h,
            y=w,
        )
        return self.to_out(out)


# x = torch.randn(1, 3, 64, 64)
# peg = ScalableImgSelfAttention(3)
# out = peg(x)
# print(out.shape)



from torch import nn


class NormalizationFractral(nn.Module):
    """
    A module that performs normalization using fractal layers.

    Args:
        dim (int): The input dimension.
        eps (float, optional): A small value added to the denominator for numerical stability. Default is 1e-8.
        fi (int, optional): The number of fractal layers. Default is 4.
        *args: Variable length argument list.
        **kwargs: Arbitrary keyword arguments.

    Attributes:
        eps (float): A small value added to the denominator for numerical stability.
        fi (int): The number of fractal layers.
        norm (nn.LayerNorm): The initial normalization layer.
        norm_i (nn.LayerNorm): Fractal normalization layers.

    """

    def __init__(
        self, dim: int, eps=1e-8, fi: int = 4, *args, **kwargs  # Fractal index
    ):
        super().__init__(*args, **kwargs)
        self.eps = eps
        self.fi = fi

        self.norm = nn.LayerNorm(dim)

        for i in range(fi):
            setattr(self, f"norm_{i}", nn.LayerNorm(dim))

    def forward(self, x):
        """
        Forward pass of the NormalizationFractral module.

        Args:
            x (torch.Tensor): The input tensor.

        Returns:
            torch.Tensor: The normalized output tensor.

        """
        x = self.norm(x)

        for i in range(self.fi):
            norm = getattr(self, f"norm_{i}")
            x = norm(x)

        return x



import math

import torch
import torch.nn as nn
import torch.nn.functional as F


class QUIK(nn.Module):
    """
    QUICK Linear layer

    Args:
        in_features: number of input features
        out_features: number of output features
        bias: whether to use bias

    Returns:
        x + proj(x) where proj is a small MLP


    Usage:
    import torch

    # Initialize the QUIK module
    quik = QUIK(in_features=784, out_features=10)

    # Create some dummy data, e.g., simulating a batch of MNIST images
    data = torch.randn(10, 784)

    # Run the data through the network
    output = quik(data)
    print(output)


    """

    def __init__(self, in_features, out_features, bias=True):
        super().__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.weight = nn.Parameter(torch.Tensor(out_features, in_features))
        self.bias = nn.Parameter(torch.Tensor(out_features)) if bias else None
        self.quantize_range = (
            8  # Assuming 4-bit quantization, so range is [-8, 7]
        )
        self.half_range = self.quantize_range // 2

        nn.init.kaiming_uniform_(self.weight, a=math.sqrt(5))
        if self.bias is not None:
            fan_in, _ = nn.init._calculate_fan_in_and_fan_out(self.weight)
            bound = 1 / math.sqrt(fan_in)
            nn.init.uniform_(self.bias, -bound, bound)

    def quantize(self, input_tensor):
        """
        Args:
            input_tensor: input tensor to be quantized

        Returns:
            output_tensor: quantized tensor
            zero_act: zero-point
            scale_act: scale factor



        """
        zero_act, scale_act = self.find_zero_scale(input_tensor)
        output_tensor = (
            ((input_tensor - zero_act) / scale_act - self.half_range)
            .clamp(-self.half_range, self.half_range - 1)
            .int()
        )
        return output_tensor, zero_act, scale_act

    def dequantize(self, input_tensor, zero_act, scale_act, scale_weight):
        """
        Dequantize the input tensor

        Args:
            input_tensor: input tensor to be dequantized
            zero_act: zero-point
            scale_act: scale factor
            scale_weight: scale factor for weight

        Returns:

            output_tensor: dequantized tensor


        """
        weights_reduced = self.weight.sum(dim=1)
        x = input_tensor.float() * scale_act * scale_weight
        shift = (
            zero_act + self.half_range * scale_act
        ) * weights_reduced.unsqueeze(-1)
        output_tensor = x + shift
        return output_tensor

    def find_zero_scale(self, input_tensor):
        """
        Find zero_scale


        Args:
            input_tensor: input tensor to be quantized

        Returns:
            zero_act: zero-point
            scale_act: scale factor



        """
        zero_act = input_tensor.min()
        scale_act = (input_tensor.max() - zero_act) / (2 * self.half_range)
        return zero_act, scale_act

    def forward(self, x):
        """
        Forward pass of the QUIK layer

        Args:
            x: input tensor

        Returns:
            output_tensor: output tensor after forward pass

        """
        # Quantize activations
        x_quant, zero_act, scale_act = self.quantize(x)

        # Quantized weight multiplication
        quant_weight = self.quantize(self.weight)[0]
        result = F.linear(
            x_quant.float(), quant_weight.float(), self.bias
        )  # Assuming INT32 multiplication result

        # Dequantization
        scale_weight = (self.weight.max() - self.weight.min()) / (
            2 * self.half_range
        )
        return self.dequantize(result, zero_act, scale_act, scale_weight)



import math
from typing import List, Dict, Optional, Tuple, Callable

import torch
from torch import Tensor
from torch.optim import Optimizer


class FastAdaptiveOptimizer(Optimizer):
    """
    FastAdaptiveOptimizer implements a novel optimization algorithm that aims to be
    faster than Adam while maintaining accuracy and potentially finding local minima more quickly.

    It combines elements from Adam, RMSprop, and SGD with momentum, while adding
    features like warmup and the option to switch between adaptive and non-adaptive modes.

    Args:
        params (iterable): iterable of parameters to optimize or dicts defining
            parameter groups
        lr (float, optional): learning rate (default: 1e-3)
        betas (Tuple[float, float], optional): coefficients used for computing
            running averages of gradient and its square (default: (0.9, 0.999))
        eps (float, optional): term added to the denominator to improve
            numerical stability (default: 1e-8)
        momentum (float, optional): momentum factor (default: 0.9)
        adaptive_lr (bool, optional): whether to use adaptive learning rate
            (default: True)
        warmup_steps (int, optional): number of warmup steps (default: 1000)
        weight_decay (float, optional): weight decay (L2 penalty) (default: 0)
        amsgrad (bool, optional): whether to use the AMSGrad variant (default: False)
    """

    def __init__(
        self,
        params: List[Tensor],
        lr: float = 1e-3,
        betas: Tuple[float, float] = (0.9, 0.999),
        eps: float = 1e-8,
        momentum: float = 0.9,
        adaptive_lr: bool = True,
        warmup_steps: int = 1000,
        weight_decay: float = 0,
        amsgrad: bool = False,
    ):
        if not 0.0 <= lr:
            raise ValueError(f"Invalid learning rate: {lr}")
        if not 0.0 <= eps:
            raise ValueError(f"Invalid epsilon value: {eps}")
        if not 0.0 <= betas[0] < 1.0:
            raise ValueError(f"Invalid beta parameter at index 0: {betas[0]}")
        if not 0.0 <= betas[1] < 1.0:
            raise ValueError(f"Invalid beta parameter at index 1: {betas[1]}")
        if not 0.0 <= momentum < 1.0:
            raise ValueError(f"Invalid momentum value: {momentum}")
        if not 0.0 <= weight_decay:
            raise ValueError(f"Invalid weight_decay value: {weight_decay}")

        defaults = dict(
            lr=lr,
            betas=betas,
            eps=eps,
            momentum=momentum,
            adaptive_lr=adaptive_lr,
            warmup_steps=warmup_steps,
            weight_decay=weight_decay,
            amsgrad=amsgrad,
        )
        super(FastAdaptiveOptimizer, self).__init__(params, defaults)

    def __setstate__(self, state: Dict) -> None:
        super(FastAdaptiveOptimizer, self).__setstate__(state)
        for group in self.param_groups:
            group.setdefault("amsgrad", False)

    @torch.no_grad()
    def step(
        self, closure: Optional[Callable[[], float]] = None
    ) -> Optional[float]:
        """
        Performs a single optimization step.

        Args:
            closure (callable, optional): A closure that reevaluates the model
                and returns the loss.

        Returns:
            Optional[float]: The loss if closure is provided, else None.
        """
        loss = None
        if closure is not None:
            with torch.enable_grad():
                loss = closure()

        for group in self.param_groups:
            params_with_grad = []
            grads = []
            exp_avgs = []
            exp_avg_sqs = []
            momentum_buffers = []
            max_exp_avg_sqs = []
            state_steps = []

            for p in group["params"]:
                if p.grad is not None:
                    params_with_grad.append(p)
                    if p.grad.is_sparse:
                        raise RuntimeError(
                            "FastAdaptiveOptimizer does not support sparse gradients"
                        )
                    grads.append(p.grad)

                    state = self.state[p]
                    # Lazy state initialization
                    if len(state) == 0:
                        state["step"] = 0
                        state["exp_avg"] = torch.zeros_like(
                            p, memory_format=torch.preserve_format
                        )
                        state["exp_avg_sq"] = torch.zeros_like(
                            p, memory_format=torch.preserve_format
                        )
                        state["momentum_buffer"] = torch.zeros_like(
                            p, memory_format=torch.preserve_format
                        )
                        if group["amsgrad"]:
                            state["max_exp_avg_sq"] = torch.zeros_like(
                                p, memory_format=torch.preserve_format
                            )

                    exp_avgs.append(state["exp_avg"])
                    exp_avg_sqs.append(state["exp_avg_sq"])
                    momentum_buffers.append(state["momentum_buffer"])

                    if group["amsgrad"]:
                        max_exp_avg_sqs.append(state["max_exp_avg_sq"])

                    state["step"] += 1
                    state_steps.append(state["step"])

            beta1, beta2 = group["betas"]
            self._update_params(
                params_with_grad,
                grads,
                exp_avgs,
                exp_avg_sqs,
                max_exp_avg_sqs,
                momentum_buffers,
                state_steps,
                group["amsgrad"],
                beta1,
                beta2,
                group["lr"],
                group["weight_decay"],
                group["eps"],
                group["momentum"],
                group["adaptive_lr"],
                group["warmup_steps"],
            )

        return loss

    def _update_params(
        self,
        params: List[Tensor],
        grads: List[Tensor],
        exp_avgs: List[Tensor],
        exp_avg_sqs: List[Tensor],
        max_exp_avg_sqs: List[Tensor],
        momentum_buffers: List[Tensor],
        state_steps: List[int],
        amsgrad: bool,
        beta1: float,
        beta2: float,
        lr: float,
        weight_decay: float,
        eps: float,
        momentum: float,
        adaptive_lr: bool,
        warmup_steps: int,
    ) -> None:
        """
        Function to update parameters.
        """
        for i, param in enumerate(params):
            grad = grads[i]
            exp_avg = exp_avgs[i]
            exp_avg_sq = exp_avg_sqs[i]
            momentum_buffer = momentum_buffers[i]
            step = state_steps[i]

            if weight_decay != 0:
                grad = grad.add(param, alpha=weight_decay)

            # Decay the first and second moment running average coefficient
            exp_avg.mul_(beta1).add_(grad, alpha=1 - beta1)
            exp_avg_sq.mul_(beta2).addcmul_(grad, grad, value=1 - beta2)

            if amsgrad:
                torch.maximum(
                    max_exp_avg_sqs[i], exp_avg_sq, out=max_exp_avg_sqs[i]
                )
                denom = max_exp_avg_sqs[i].sqrt().add_(eps)
            else:
                denom = exp_avg_sq.sqrt().add_(eps)

            bias_correction1 = 1 - beta1**step
            bias_correction2 = 1 - beta2**step

            if adaptive_lr:
                step_size = lr * math.sqrt(bias_correction2) / bias_correction1
            else:
                step_size = lr

            # Apply warmup
            if step < warmup_steps:
                step_size *= step / warmup_steps

            # Compute the update
            if adaptive_lr:
                update = exp_avg / denom
            else:
                update = exp_avg

            # Apply momentum and update parameters
            momentum_buffer.mul_(momentum).add_(update)
            param.add_(momentum_buffer, alpha=-step_size)



import re
from functools import wraps

from einops import rearrange, reduce, repeat


def check_shape(tensor, pattern, **kwargs):
    return rearrange(tensor, f"{pattern} -> {pattern}", **kwargs)


# Do many ops on a list of tensors
def _many(fn):
    @wraps(fn)
    def inner(tensors, pattern, **kwargs):
        return (fn(tensor, pattern, **kwargs) for tensor in tensors)

    return inner


# Do einops with unflattening of named dimensions
# (...flatenned) -> ...flattened


def _with_anon_dims(fn):
    @wraps(fn)
    def inner(tensor, pattern, **kwargs):
        regex = r"(\.\.\.[a-zA-Z]+)"
        matches = re.findall(regex, pattern)

        def get_anon_dim_name(t):
            return t.lstrip("...")

        dim_prefixes = tuple(map(get_anon_dim_name, matches))

        update_kwargs_dict = {}

        for prefix in dim_prefixes:
            assert (
                prefix in kwargs
            ), f"dimension list {prefix} not found in kwargs"
            dim_list = kwargs[prefix]
            assert isinstance(
                dim_list, (list, tuple)
            ), f"Dimension list {prefix} needs to be a tuple of list"
            dim_names = list(
                map(lambda ind: f"{prefix}{ind}", range(len(dim_list)))
            )
            update_kwargs_dict[prefix] = dict(zip(dim_names, dim_list))

        def sub_with_anon_dims(t):
            dim_name_prefix = get_anon_dim_name(t.groups()[0])
            return "".join(update_kwargs_dict[dim_name_prefix].keys())

        pattern_new = re.sub(regex, sub_with_anon_dims, pattern)
        return fn(tensor, pattern_new, **kwargs)

    return inner


rearrange_many = _many(rearrange)
repeat_many = _many(repeat)
reduce_many = _many(reduce)

rearrange_with_anon_dims = _with_anon_dims(rearrange)
repeat_with_anon_dims = _with_anon_dims(repeat)
reduce_with_anon_dims = _with_anon_dims(reduce)



from einops import rearrange, reduce
from torch import Tensor, nn


def img_to_text(x: Tensor, seqlen: int, dim: int, norm: bool = True):
    """
    Convert an image tensor to a text tensor.

    Args:
        x (Tensor): Input image tensor of shape (batch_size, channels, height, width).
        seqlen (int): Length of the output text sequence.
        dim (int): Dimension of the intermediate representation.
        norm (bool, optional): Whether to apply layer normalization. Defaults to True.

    Returns:
        Tensor: Output text tensor of shape (batch_size, seqlen, dim).

    Example::
        >>> x = torch.randn(2, 3, 32, 32)
        >>> x = img_to_text(x, 100, 512)
        >>> x.shape
        torch.Size([2, 100, 512])
    """
    b, c, h, w = x.shape

    img = reduce(x, "b c h w -> b c (h w)", "mean")
    img = nn.Linear(h * w, dim)(img)
    img = rearrange(img, "b c d -> b d c")
    img = nn.Linear(c, seqlen)(img)
    img = rearrange(img, "b d c -> b c d")

    if norm:
        img = nn.LayerNorm(dim)(img)

    return img



import torch
from torch import Tensor


def sparsemax(x: Tensor):
    """
    A PyTorch implementation of the sparsemax function.

    Args:
        x (torch.Tensor): The x tensor.

    Returns:
        torch.Tensor: The output of the sparsemax function.

    Example:
    >>> x = torch.tensor([1, 2, 3, 4, 5], dtype=torch.float32)
    >>> sparsemax(x)
    tensor([0., 0., 0., 1., 1.])
    """
    dim = x.dim() - 1
    number_of_logits = x.size(dim)

    x = x - torch.max(x, dim=dim, keepdim=True)[0].expand_as(x)
    zs = torch.sort(x=x, dim=dim, descending=True)[0]
    range = torch.arange(
        start=1, end=number_of_logits + 1, device=x.device
    ).view(1, -1)
    range = range.expand_as(zs)

    bound = 1 + range * zs
    cumulative_sum_zs = torch.cumsum(zs, dim)
    is_gt = torch.gt(bound, cumulative_sum_zs).type(x.type())
    k = torch.max(is_gt * range, dim, keepdim=True)[0]

    zs_sparse = is_gt * zs
    taus = (torch.sum(zs_sparse, dim, keepdim=True) - 1) / k
    taus = taus.expand_as(x)
    output = torch.max(torch.zeros_like(x), x - taus)
    return output



from torch import nn
import torch.nn.functional as F
from zeta.nn.modules.glu import GLU
from zeta.nn.modules.swiglu import SwiGLU
from typing import Optional

# from zeta.experimental.triton.triton_modules.linear_proj import LinearTriton


class ReluSquared(nn.Module):
    def forward(self, x):
        return F.relu(x) ** 2


def exists(val):
    return val is not None


def default(val, default_val):
    return default_val if val is None else val


def init_zero_(layer):
    nn.init.constant_(layer.weight, 0.0)
    if exists(layer.bias):
        nn.init.constant_(layer.bias, 0.0)


class FeedForward(nn.Module):
    def __init__(
        self,
        dim: Optional[int] = None,
        dim_out: Optional[int] = None,
        mult: Optional[int] = 4,
        glu: Optional[bool] = False,
        glu_mult_bias: Optional[bool] = False,
        swish: Optional[bool] = False,
        relu_squared: Optional[bool] = False,
        post_act_ln: Optional[bool] = False,
        dropout: Optional[float] = 0.0,
        no_bias: Optional[bool] = False,
        zero_init_output: Optional[bool] = False,
        custom_act: Optional[nn.Module] = None,
        swiglu: Optional[bool] = False,
        triton_kernels_on: bool = False,
    ):
        """
        FeedForward module that applies a series of linear transformations and activations.

        Args:
            dim (int): Input dimension.
            dim_out (int, optional): Output dimension. Defaults to None.
            mult (int, optional): Multiplier for the inner dimension. Defaults to 4.
            glu (bool, optional): Whether to use Gated Linear Units (GLU). Defaults to False.
            glu_mult_bias (bool, optional): Whether to use bias in the GLU operation. Defaults to False.
            swish (bool, optional): Whether to use Swish activation. Defaults to False.
            relu_squared (bool, optional): Whether to use squared ReLU activation. Defaults to False.
            post_act_ln (bool, optional): Whether to apply Layer Normalization after the activation. Defaults to False.
            dropout (float, optional): Dropout probability. Defaults to 0.0.
            no_bias (bool, optional): Whether to use bias in the linear transformations. Defaults to False.
            zero_init_output (bool, optional): Whether to initialize the last linear layer to 0. Defaults to False.
            custom_act (nn.Module, optional): Custom activation module. Defaults to None.
            swiglu (bool, optional): Whether to use SwiGLU activation. Defaults to False.
        """
        super().__init__()
        self.dim = dim
        self.dim_out = dim_out
        self.mult = mult
        self.glu = glu
        self.glu_mult_bias = glu_mult_bias
        self.swish = swish
        self.relu_squared = relu_squared
        self.post_act_ln = post_act_ln
        self.dropout = dropout
        self.no_bias = no_bias
        self.zero_init_output = zero_init_output
        self.custom_act = custom_act
        self.swiglu = swiglu
        self.triton_kernels_on = triton_kernels_on

        inner_dim = int(dim * mult)
        dim_out = default(dim_out, dim)

        if relu_squared:
            activation = ReluSquared()
        elif swish:
            activation = nn.SiLU()
        elif custom_act is not None:
            activation = custom_act
        elif swiglu:
            activation = SwiGLU()
        else:
            activation = nn.GELU()

        if glu:
            project_in = GLU(
                dim, inner_dim, activation, mult_bias=glu_mult_bias
            )
        # elif triton_kernels_on is True:
        #     project_in = nn.Sequential(
        #         LinearTriton(dim, inner_dim, bias=no_bias), activation
        #     )
        else:
            project_in = nn.Sequential(
                nn.Linear(dim, inner_dim, bias=not no_bias), activation
            )

        if post_act_ln:
            self.ff = nn.Sequential(
                project_in,
                nn.LayerNorm(inner_dim),
                nn.Dropout(dropout),
                nn.Linear(inner_dim, dim_out, bias=no_bias),
            )
        else:
            self.ff = nn.Sequential(
                project_in,
                nn.Dropout(dropout),
                nn.Linear(inner_dim, dim_out, bias=not no_bias),
            )

        # init last linear layer to 0
        if zero_init_output:
            init_zero_(self.ff[-1])

    def forward(self, x):
        """
        Forward pass of the feedforward network

        Args:
            x (torch.Tensor): Input tensor

        Returns:
            torch.Tensor: Output tensor
        """
        return self.ff(x)



import torch
from torch import nn


class SPAct(nn.Module):
    def __init__(self, alpha: float = 0.5):
        """
        Initializes the SPAct module.

        Args:
            alpha (float): The weight parameter for the linear combination of the input and the hyperbolic tangent output.
        """
        super().__init__()
        self.alpha = alpha

    def forward(self, x):
        """
        Performs the forward pass of the SPAct module.

        Args:
            x (torch.Tensor): The input tensor.

        Returns:
            torch.Tensor: The output tensor after applying the SPAct function.
        """
        return self.alpha * x + (1 - self.alpha) * torch.tanh(x)


# x = torch.randn(1, 3)

# model = SPAct()

# out = model(x)
# print(out)



import torch
import torch.nn as nn

if torch.cuda.is_available():
    try:
        import triton
        import triton.language as tl
    except ImportError:
        print(
            "Triton is not installed. Please install it using `pip install"
            " triton`."
        )


@triton.jit
def linear_projection_kernel(
    X, W, Y, M, N, K, stride_x, stride_w, stride_y, BLOCK_SIZE: tl.constexpr
):
    # Compute indices
    row_idx = tl.program_id(0)
    col_idx = tl.program_id(1)

    # Offsets for X, W, and Y
    x_off = row_idx * stride_x
    w_off = col_idx * stride_w
    y_off = row_idx * stride_y + col_idx

    # Dot product
    acc = tl.zeros((), dtype=tl.float32)
    for k in range(K):
        acc += tl.load(X + x_off + k) * tl.load(W + w_off + k)
    tl.store(Y + y_off, acc)


class LinearTriton(nn.Module):
    """
    A custom linear module implemented using Triton.

    Args:
        in_features (int): Number of input features.
        out_features (int): Number of output features.
        bias (bool, optional): If set to True, the module has a learnable bias. Default is True.
    """

    def __init__(self, in_features, out_features, bias=True):
        super(LinearTriton, self).__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.weight = nn.Parameter(torch.randn(out_features, in_features))
        if bias:
            self.bias = nn.Parameter(torch.randn(out_features))
        else:
            self.register_parameter("bias", None)

    def forward(self, x):
        """
        Performs a forward pass through the linear module.

        Args:
            x (torch.Tensor): Input tensor of shape (batch_size, in_features).

        Returns:
            torch.Tensor: Output tensor of shape (batch_size, out_features).
        """
        # Prepare the output tensor
        output = torch.empty(
            x.shape[0], self.out_features, device=x.device, dtype=x.dtype
        )

        # Grid and block dimensions
        grid = (x.shape[0], self.out_features)
        block = 128  # Example block size

        # Launch the Triton kernel
        linear_projection_kernel[grid](
            x,
            self.weight,
            output,
            x.shape[0],
            self.out_features,
            self.in_features,
            x.stride(0),
            self.weight.stride(0),
            output.stride(0),
            block,
        )

        # Add bias if present
        if self.bias is not None:
            output += self.bias.unsqueeze(0)  # Broadcasting the bias
        return output


# # Example usage
# model = LinearTriton(128, 64).cuda()
# input_tensor = torch.randn(1, 10, 128).cuda()
# output_tensor = model(input_tensor)
# print(output_tensor.shape)  # Should be torch.Size([10, 64])



import torch
import torch.nn as nn
from einops import rearrange

from zeta.ops.einops_poly import rearrange_many


class SpatialLinearAttention(nn.Module):
    """
    Spatial Linear Attention module.

    Args:
        dim (int): Input dimension. Defaults to None.
        heads (int): Number of attention heads. Defaults to 4.
        dim_head (int): Dimension of each attention head. Defaults to 32.
    """

    def __init__(self, dim: int = None, heads: int = 4, dim_head: int = 32):
        super().__init__()
        self.scale = dim_head**-0.5
        self.heads = heads
        hidden_dim = dim_head * heads

        self.to_qkv = nn.Conv2d(dim, hidden_dim * 3, 1, bias=False)
        self.to_out = nn.Conv2d(hidden_dim, dim, 1)

    def forward(self, x):
        """
        Forward pass of the Spatial Linear Attention module.

        Args:
            x (torch.Tensor): Input tensor of shape (batch_size, channels, frames, height, width).

        Returns:
            torch.Tensor: Output tensor of shape (batch_size, channels, frames, height, width).
        """
        b, c, f, h, w = x.shape
        x = rearrange(x, "b c f h w -> (b f) c h w")

        qkv = self.to_qkv(x).chunk(3, dim=1)
        q, k, v = rearrange_many(
            qkv, "b (h c) x y -> b h c (x y)", h=self.heads
        )

        q = q.softmax(dim=-2)
        k = k.softmax(dim=-1)

        q = q * self.scale
        context = torch.einsum("b h d n, b h e n -> b h d e", k, v)

        out = torch.einsum("b h d e, b h d n -> b h e n", context, q)
        out = rearrange(
            out, "b h c (x y) -> b (h c) x y", h=self.heads, x=h, y=w
        )
        out = self.to_out(out)

        return rearrange(out, "(b f) c h w -> b c f h w", b=b)



import torch
import torch.nn.functional as F
from torch import Tensor, nn


class WSConv2d(nn.Conv2d):
    """
    Weight Standardized Convolutional 2D Layer.

    This class inherits from `nn.Conv2d` and adds weight standardization to the convolutional layer.
    It normalizes the weights of the convolutional layer to have zero mean and unit variance along
    the channel dimension. This helps in stabilizing the training process and improving generalization.

    Args:
        in_channels (int): Number of input channels.
        out_channels (int): Number of output channels.
        kernel_size (int): Size of the convolutional kernel.
        stride (float, optional): Stride of the convolution. Default is 1.
        padding (int or tuple, optional): Padding added to the input. Default is 0.
        dilation (int, optional): Spacing between kernel elements. Default is 1.
        groups (int, optional): Number of blocked connections from input channels to output channels. Default is 1.
        bias (bool, optional): If True, adds a learnable bias to the output. Default is True.
        padding_mode (str, optional): Type of padding. Default is "zeros".
    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int,
        stride: float = 1,
        padding=0,
        dilation: int = 1,
        groups: int = 1,
        bias: bool = True,
        padding_mode: str = "zeros",
    ):
        super().__init__(
            in_channels,
            out_channels,
            kernel_size,
            stride,
            padding,
            dilation,
            groups,
            bias,
            padding_mode,
        )

        nn.init.xavier_normal_(self.weight)

        # Params
        self.gain = nn.Parameter(torch.ones(self.out_channels, 1, 1, 1))
        self.register_buffer(
            "eps", torch.tensor(1e-4, requires_grad=False), persistent=False
        )
        self.register_buffer(
            "fan_in",
            torch.tensor(
                self.weight.shape[1:].numel(), requires_grad=False
            ).type_as(self.weight),
            persistent=False,
        )

    def standardized_weights(self):
        mean = torch.mean(self.weight, axis=[1, 2, 3], keepdims=True)
        var = torch.var(self.weight, axis=[1, 2, 3], keepdims=True)
        scale = torch.rsqrt(torch.maximum(var * self.fan_in, self.eps))
        return (self.weight - mean) * scale * self.gain

    def forward(self, x: Tensor):
        return F.conv2d(
            input=x,
            weight=self.standardized_weights(),
            bias=self.bias,
            stride=self.stride,
            padding=self.padding,
            dilation=self.dilation,
            groups=self.groups,
        )



from torch import nn


class SubLN(nn.Module):
    """
    SubLN (Subtraction & Layer Normalization) module.

    This module computes the subln function:
    subln(x) = x + fout(LN(fin(LN(x))))

    Parameters:
    -----------
    dim: int
        The number of expected features in the input x
    γ: float, optional
        Gain value for weight initialization. Default is 1.0.


    # Example usage

    # Usage example:
    import torch
    from zeta.nn.modules import SubLN

    model = SubLN(dim=512)
    x = torch.randn(10, 512)
    out = model(x)
    print(out)

    """

    def __init__(self, dim, γ=1.0):
        super().__init__()

        # Define necessary layers and operations
        self.LN1 = nn.LayerNorm(dim)
        self.fin = nn.Linear(dim, dim)  # Example layer for fin
        self.fout = nn.Linear(dim, dim)  # Example layer for fout
        self.LN2 = nn.LayerNorm(dim)

        # Weight initialization
        self._initialize_weights(γ)

    def forward(self, x):
        """
        Forward pass for the subln function.

        Parameters:
        -----------
        x : torch.Tensor
            Input tensor of shape [batch_size, dim]

        Returns:
        --------
        torch.Tensor
            Output tensor of shape [batch_size, dim]

        """
        return x + self.fout(self.LN2(self.fin(self.LN1(x))))

    def _initialize_weights(self, γ):
        """
        Initialize weights of the module.

        Parameters:
        -----------
        γ : float
            Gain value for weight initialization.

        """
        for name, param in self.named_parameters():
            if "weight" in name:
                if name in ["fin.weight", "fout.weight", "out_proj.weight"]:
                    nn.init.xavier_normal_(param, gain=γ)
                elif name in ["q_proj.weight", "k_proj.weight"]:
                    nn.init.xavier_normal_(param, gain=1)
            elif "bias" in name:
                nn.init.zeros_(param)



import functools
import time
from torch import Tensor, nn
from loguru import logger
from typing import Dict, Any, Callable
import torch.cuda as cuda
from typing import Union


def verbose_execution(
    log_params: bool = False,
    log_gradients: bool = False,
    log_memory: bool = True,
):
    """
    A decorator that adds verbose logging to a PyTorch model's execution.

    Args:
        log_params (bool): Whether to log parameter shapes. Defaults to False.
        log_gradients (bool): Whether to log gradient information. Defaults to False.
        log_memory (bool): Whether to log memory usage. Defaults to True.

    Returns:
        Callable: A decorator function.
    """

    def decorator(model_class: Callable) -> Callable:
        @functools.wraps(model_class)
        def wrapper(*args, **kwargs) -> nn.Module:
            model = model_class(*args, **kwargs)
            return VerboseExecution(
                model, log_params, log_gradients, log_memory
            )

        return wrapper

    return decorator


class VerboseExecution(nn.Module):
    def __init__(
        self,
        model: nn.Module,
        log_params: bool,
        log_gradients: bool,
        log_memory: bool,
    ):
        super().__init__()
        self.model = model
        self.log_params = log_params
        self.log_gradients = log_gradients
        self.log_memory = log_memory
        self._register_hooks()

    def _register_hooks(self):
        for name, module in self.model.named_modules():
            module.__name__ = name
            module.register_forward_hook(self._create_forward_hook(name))
            if self.log_gradients:
                module.register_full_backward_hook(
                    self._create_backward_hook(name)
                )

    def _create_forward_hook(self, name: str):
        def hook(
            module: nn.Module,
            input: Union[Tensor, tuple],
            output: Union[Tensor, tuple],
        ):
            log_dict: Dict[str, Any] = {
                "module": name,
                "input_shape": self._get_shape(input),
                "output_shape": self._get_shape(output),
            }

            if self.log_params:
                log_dict["parameters"] = {
                    param_name: tuple(param.shape)
                    for param_name, param in module.named_parameters()
                }

            if self.log_memory:
                log_dict["memory"] = self._get_memory_usage()

            logger.info(log_dict)

        return hook

    def _create_backward_hook(self, name: str):
        def hook(
            module: nn.Module,
            grad_input: Union[Tensor, tuple],
            grad_output: Union[Tensor, tuple],
        ):
            log_dict: Dict[str, Any] = {
                "module": name,
                "grad_input_shape": self._get_shape(grad_input),
                "grad_output_shape": self._get_shape(grad_output),
            }
            logger.info(f"Backward pass: {log_dict}")

        return hook

    @staticmethod
    def _get_shape(tensor_or_tuple: Union[Tensor, tuple]) -> Union[tuple, list]:
        if isinstance(tensor_or_tuple, tuple):
            return [
                VerboseExecution._get_shape(t)
                for t in tensor_or_tuple
                if t is not None
            ]
        elif isinstance(tensor_or_tuple, Tensor):
            return tuple(tensor_or_tuple.shape)
        else:
            return None

    @staticmethod
    def _get_memory_usage() -> Dict[str, float]:
        return {
            "allocated": cuda.memory_allocated() / 1024**2,
            "cached": cuda.memory_reserved() / 1024**2,
        }

    def forward(self, *args, **kwargs) -> Any:
        start_time = time.time()
        logger.info(
            f"Model input shapes: {[self._get_shape(arg) for arg in args]}"
        )

        output = self.model(*args, **kwargs)

        end_time = time.time()
        logger.info(f"Model output shape: {self._get_shape(output)}")
        logger.info(f"Forward pass time: {end_time - start_time:.4f} seconds")

        if self.log_memory:
            logger.info(f"GPU memory usage: {self._get_memory_usage()}")

        return output



import torch
import torch.nn.functional as F
from beartype import beartype
from beartype.typing import Optional
from einops import pack, rearrange, reduce, unpack
from torch import Tensor, nn
from torch.nn import Module

# helper


def exists(v):
    return v is not None


def default(v, d):
    return v if exists(v) else d


def identity(t):
    return t


def divisible_by(num, den):
    return (num % den) == 0


def pack_one(t, pattern):
    return pack([t], pattern)


def unpack_one(t, ps, pattern):
    return unpack(t, ps, pattern)[0]


def is_odd(n):
    return not divisible_by(n, 2)


def cast_tuple(t, length=1):
    return t if isinstance(t, tuple) else ((t,) * length)


# helper classes


def Sequential(*modules):
    modules = [*filter(exists, modules)]

    if len(modules) == 0:
        return nn.Identity()

    return nn.Sequential(*modules)


class Residual(Module):
    def __init__(self, fn):
        super().__init__()
        self.fn = fn

    def forward(self, x, **kwargs):
        return self.fn(x, **kwargs) + x


# adaptive conv from Karras et al. Stylegan2
# for conditioning on latents


class AdaptiveConv3DMod(Module):
    """
    Adaptive convolutional layer, with support for spatial modulation

    Args:
        dim: input channels
        spatial_kernel: spatial kernel size
        time_kernel: temporal kernel size
        dim_out: output channels
        demod: demodulate weights
        eps: epsilon for numerical stability

    Returns:
        Tensor of shape (batch, channels, time, height, width)

    Examples:
    >>> x = torch.randn(1, 512, 4, 4, 4)
    >>> mod = torch.randn(1, 512)
    >>> layer = AdaptiveConv3DMod(512, 3, 3)
    >>> out = layer(x, mod)
    >>> out.shape


    """

    @beartype
    def __init__(
        self,
        dim,
        *,
        spatial_kernel,
        time_kernel,
        dim_out=None,
        demod=True,
        eps=1e-8,
    ):
        super().__init__()
        dim_out = default(dim_out, dim)

        self.eps = eps

        assert is_odd(spatial_kernel)
        assert is_odd(time_kernel)

        self.spatial_kernel = spatial_kernel
        self.time_kernel = time_kernel

        self.padding = (
            *((spatial_kernel // 2,) * 4),
            *((time_kernel // 2,) * 2),
        )
        self.weights = nn.Parameter(
            torch.randn(
                (dim_out, dim, time_kernel, spatial_kernel, spatial_kernel)
            )
        )

        self.demod = demod

        nn.init.kaiming_normal_(
            self.weights, a=0, mode="fan_in", nonlinearity="selu"
        )

    def forward(self, fmap, mod: Optional[Tensor] = None):
        """
        notation

        b - batch
        n - convs
        o - output
        i - input
        k - kernel
        """

        b = fmap.shape[0]

        # prepare weights for modulation

        weights = self.weights

        # do the modulation, demodulation, as done in stylegan2

        mod = rearrange(mod, "b i -> b 1 i 1 1 1")

        weights = weights * (mod + 1)

        if self.demod:
            inv_norm = (
                reduce(weights**2, "b o i k0 k1 k2 -> b o 1 1 1 1", "sum")
                .clamp(min=self.eps)
                .rsqrt()
            )
            weights = weights * inv_norm

        fmap = rearrange(fmap, "b c t h w -> 1 (b c) t h w")

        weights = rearrange(weights, "b o ... -> (b o) ...")

        fmap = F.pad(fmap, self.padding)
        fmap = F.conv3d(fmap, weights, groups=b)

        return rearrange(fmap, "1 (b o) ... -> b o ...", b=b)



import torch
from torch import Tensor


# Define the TensorShape class
class TensorShape(Tensor):
    """
    Represents the shape of a tensor.

    Args:
        data (array-like): The data of the tensor.
        shape_string (str): The string representation of the shape.

    Attributes:
        shape_string (str): The string representation of the shape.
        shape_dict (dict): A dictionary mapping dimensions to sizes.

    Raises:
        ValueError: If the shape string does not match the actual shape.

    Example:
        >>> data = [1, 2, 3, 4]
        >>> shape_string = "2 2"
        >>> tensor_shape = TensorShape(data, shape_string)
        >>> print(tensor_shape)
        TensorShape(shape_string='2 2', actual_shape=(2, 2))
    """

    def __new__(cls, data, shape_string):
        instance = torch.as_tensor(data).as_subclass(cls)
        instance.shape_string = shape_string
        instance.shape_dict = cls.parse_shape_string(
            shape_string, instance.shape
        )
        return instance

    @staticmethod
    def parse_shape_string(shape_string, actual_shape):
        """
        Parses the shape string and returns a dictionary mapping dimensions to sizes.

        Args:
            shape_string (str): The string representation of the shape.
            actual_shape (tuple): The actual shape of the tensor.

        Returns:
            dict: A dictionary mapping dimensions to sizes.

        Raises:
            ValueError: If the number of dimensions in the shape string does not match the actual shape.
        """
        dimensions = shape_string.split()
        if len(dimensions) != len(actual_shape):
            raise ValueError(
                f"Shape string {shape_string} does not match actual shape {actual_shape}"
            )
        return {dim: size for dim, size in zip(dimensions, actual_shape)}

    def __repr__(self):
        return f"TensorShape(shape_string={self.shape_string}, actual_shape={super().shape})"

    @staticmethod
    def check_shape(tensor, shape_string):
        """
        Checks if the shape of the given tensor matches the specified shape string.

        Args:
            tensor (Tensor): The tensor to check the shape of.
            shape_string (str): The string representation of the expected shape.

        Raises:
            ValueError: If the shape of the tensor does not match the expected shape.
        """
        shape_dict = TensorShape.parse_shape_string(shape_string, tensor.shape)
        if tensor.shape != tuple(shape_dict.values()):
            raise ValueError(
                f"Expected shape {shape_dict}, but got {tensor.shape}"
            )


# Define a decorator for shape checking
def check_tensor_shape(shape_string: str = None):
    """
    Decorator function that checks if the shape of a tensor matches the specified shape string.

    Args:
        shape_string (str): A string representing the desired shape of the tensor.

    Returns:
        function: A decorator function that wraps the original function and performs the shape check.

    Example:
        @check_tensor_shape("B S D")
        def my_function(tensor):
            # Function implementation
            pass

        The above example will ensure that the tensor passed to `my_function` has a shape of (2, 3).
    """

    def decorator(func):
        def wrapper(*args, **kwargs):
            # Assuming the tensor is the first argument
            tensor = args[1]
            TensorShape.check_shape(tensor, shape_string)
            return func(*args, **kwargs)

        return wrapper

    return decorator


# Define a helper function to create TensorShape objects
def create_tensor(
    data: Tensor = None, shape_string: str = None, random_on: bool = False
):
    if random_on:
        data = torch.randn(data)
        return TensorShape(data, shape_string)
    else:
        return TensorShape(data, shape_string)



from torch import nn
import torch.nn.functional as F


class ReluSquared(nn.Module):
    """
    Applies the ReLU activation function and squares the output.

    Args:
        x (torch.Tensor): Input tensor.

    Returns:
        torch.Tensor: Output tensor after applying ReLU and squaring the result.
    """

    def forward(self, x):
        return F.relu(x) ** 2



import torch
from torch import nn


class SkipConnection(nn.Module):
    def __init__(self, submodule):
        super().__init__()
        self.submodule = submodule

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Forward pass of the SkipConnection module.

        Args:
            x (torch.Tensor): Input tensor.

        Returns:
            torch.Tensor: Output tensor after adding the input tensor with the submodule output.
        """
        return x + self.submodule(x)



# the best llm ever made
from torch.nn import Module

from zeta.structs.transformer import Decoder, Transformer
from zeta.structs.auto_regressive_wrapper import AutoRegressiveWrapper


class Andromeda(Module):
    """
    Andromeda is a transformer-based model architecture. It initializes with
    a Transformer and AutoRegressiveWrapper with default or user-specified parameters.
    """

    def __init__(
        self,
        num_tokens=50432,
        max_seq_len=8192,
        dim=2560,
        depth=32,
        dim_head=128,
        heads=24,
        use_abs_pos_emb=False,
        alibi_pos_bias=True,
        alibi_num_heads=12,
        rotary_xpos=True,
        attn_flash=True,
        attn_kv_heads=2,
        qk_norm=True,
        attn_qk_norm=True,
        attn_qk_norm_dim_scale=True,
    ):
        """
        Initialize the model with specified or default parameters.
        Args:
        - num_tokens: Number of tokens in the vocabulary
        - max_seq_len: Maximum sequence length
        - dim: Dimension of the model
        - depth: Depth of the model
        - dim_head: Dimension of the model head
        - heads: Number of heads
        - use_abs_pos_emb: Whether to use absolute position embedding
        - alibi_pos_bias: Alibi position bias
        - alibi_num_heads: Number of alibi heads
        - rotary_xpos: Rotary position
        - attn_flash: Attention flash
        - deepnorm: Deep normalization
        - shift_tokens: Number of tokens to shift
        - attn_one_kv_head: Attention one key/value head
        - qk_norm: Query-key normalization
        - attn_qk_norm: Attention query-key normalization
        - attn_qk_norm_dim_scale: Attention query-key normalization dimension scale
        - embedding_provider: Embedding provider module
        """
        super().__init__()

        try:
            self.Andromeda = Transformer(
                num_tokens=num_tokens,
                max_seq_len=max_seq_len,
                use_abs_pos_emb=use_abs_pos_emb,
                attn_layers=Decoder(
                    dim=dim,
                    depth=depth,
                    dim_head=dim_head,
                    heads=heads,
                    alibi_pos_bias=alibi_pos_bias,
                    alibi_num_heads=alibi_num_heads,
                    rotary_xpos=rotary_xpos,
                    attn_flash=attn_flash,
                    attn_kv_heads=attn_kv_heads,
                    qk_norm=qk_norm,
                    attn_qk_norm=attn_qk_norm,
                    attn_qk_norm_dim_scale=attn_qk_norm_dim_scale,
                ),
            )

            self.decoder = AutoRegressiveWrapper(self.Andromeda)

        except Exception as e:
            print("Failed to initialize Andromeda: ", e)
            raise

    def forward(self, text_tokens, **kwargs):
        """
        Forward pass through the model. It expects the input text_tokens.
        Args:
        - text_tokens: Input tokens
        - kwargs: Other arguments
        Returns:
        - output from the decoder
        """
        try:
            model_input = self.decoder.forward(text_tokens)[0]
            return self.decoder(model_input, padded_x=model_input[0])
        except Exception as e:
            print("Failed in forward method: ", e)
            raise



# Copyright (c) 2022 Agora
# Licensed under The MIT License [see LICENSE for details]

# Copyright (c) Facebook, Inc. and its affiliates. All rights reserved.
#
# This source code is licensed under the BSD license found in the
# LICENSE file in the root directory of this source tree.

# NOTE: This is a mirror of the code in
# https://github.com/facebookresearch/fairscale/tree/master/fairscale/nn/moe

import logging
import time
from typing import Any, Tuple, cast

import torch
import torch.distributed as dist
from torch import Tensor
from torch.nn import Module, ModuleList

from .global_groups import get_all2all_group, get_moe_group

try:
    from fairseq.modules.moe import MOELayer

    has_fairseq = True
    Base = MOELayer
except ModuleNotFoundError:
    Base = Module
    has_fairseq = False

try:
    # To enable Tutel MoE optimizations:
    # python3 -m pip install --user --upgrade
    # git+https://github.com/Agora/tutel@v0.1.x
    from tutel import moe as tutel_moe

    has_tutel, fused_cumsum_sub_one = True, tutel_moe.fast_cumsum_sub_one
except ModuleNotFoundError:
    has_tutel, fused_cumsum_sub_one = (
        False,
        lambda mask: torch.cumsum(mask, dim=0) - 1,
    )

logger = logging.getLogger(__name__)

# einsum dimensions: (g)roup, (s)equence, (e)xpert, (m)odel, (c)apacity
# See https://arxiv.org/pdf/2006.16668.pdf for details.


# Based on https://github.com/pytorch/pytorch/pull/40762
class _AllToAll(torch.autograd.Function):
    @staticmethod
    def forward(
        ctx: Any, group: dist.ProcessGroup, input: Tensor
    ) -> Tensor:  # type: ignore
        ctx.group = group
        input = input.contiguous()
        output = torch.empty_like(input)
        if torch.distributed.is_initialized():
            dist.all_to_all_single(output, input, group=group)
        else:
            assert group is None
            output = input
        return output

    @staticmethod
    def backward(ctx: Any, *grad_output: Tensor) -> Tuple[None, Tensor]:
        return (None, _AllToAll.apply(ctx.group, *grad_output))


class MOELayer(Base):
    """MOELayer module which implements MixtureOfExperts as described in Gshard_.
    ::

        gate = Top2Gate(model_dim, num_experts)
        moe = MOELayer(gate, expert)
        output = moe(input)
        l_aux = moe.l_aux

    .. Gshard_: https://arxiv.org/pdf/2006.16668.pdf

    Args:
        gate (torch.nn.Module):
            gate network
        expert (torch.nn.Module):
            expert network
    """

    def __init__(self, gate, experts, args):
        if has_fairseq:
            super(Base, self).__init__()
        else:
            super().__init__()
        self.gate = gate
        if isinstance(experts, ModuleList):
            self.experts = cast(ModuleList, experts)
        else:
            self.experts = ModuleList([experts])
        _, self.expert_group = get_moe_group(args.moe_expert_count)
        self.all2all_group = get_all2all_group(args.moe_expert_count)
        self.world_size = dist.get_world_size(group=self.expert_group)
        self.all2all_size = dist.get_world_size(group=self.all2all_group)
        for p in experts.parameters():
            p.expert = True  # type: ignore
        self.num_local_experts = len(self.experts)
        self.args = args
        self.in_generation = False
        self.a2a_cuda_event_intervals = []
        self.a2a_cpu_time_ms = 0.0

    def forward(
        self, *input: Tensor, input_padding_mask=None, **kwargs: Any
    ) -> Tensor:
        assert len(input) == 1, "only single input Tensor supported"
        input = input[0]
        assert (
            len(input.shape) == 3
        ), "input Tensor must have dimensions: (s)equence, (t)oken, (m)odel"
        if input_padding_mask is not None:
            assert (
                len(input_padding_mask.shape) == 2
            ), "input Tensor must have dimensions: (s)equence, (t)oken"
            assert input_padding_mask.shape[0] == input.shape[0]
            assert input_padding_mask.shape[1] == input.shape[1]
        # assert input.shape[0] % len(self.experts) == 0, "num tokens must be order of number of local experts"

        # Implement Algorithm 2 from GShard paper.
        dim = input.shape[2]
        # Pad to expected batch size
        input_shape = list(input.shape)
        expected_bsz = (
            getattr(self.args, "batch_size", 0)
            if self.training
            else getattr(self.args, "batch_size_valid", 0)
        )
        # This indicates that --batch-size or --max-sentences is not specified
        if expected_bsz is None:
            expected_bsz = 0
        # Note: Padding is not necessary at generation time at present
        # because all DDP workers process the same batch. Also, batch size at generation time
        # can be different from that present in the checkpoint state
        if (
            not self.in_generation
            and expected_bsz != 0
            and input_shape[0] != expected_bsz
        ):
            logger.warning(
                "padding batch with unexpected size"
                f" {input_shape[0]} (expected: {expected_bsz})"
            )
            assert (
                input_shape[0] < expected_bsz
            ), f"{input_shape[0]} < {expected_bsz}"
            padded_input = torch.zeros(
                (expected_bsz, input_shape[1], input_shape[2]),
                dtype=input.dtype,
                layout=input.layout,
                device=input.device,
            )
            padded_input[: input_shape[0], :, :] = input
            input = padded_input

            padded_input_padding_mask = torch.ones(
                (
                    expected_bsz,
                    input_shape[1],
                ),
                dtype=torch.bool,
                device=input.device,
            )
            if input_padding_mask is not None:
                padded_input_padding_mask[: input_shape[0], :] = (
                    input_padding_mask
                )
            else:
                padded_input_padding_mask[: input_shape[0], :] = False
            input_padding_mask = padded_input_padding_mask

        # Reshape into S tokens by dropping sequence dimension.
        reshaped_input = input.reshape(-1, dim)
        reshaped_input_shape = reshaped_input.shape
        reshaped_input_padding_mask = (
            input_padding_mask.reshape(-1)
            if input_padding_mask is not None
            else None
        )

        # Doing padding here when --max-tokens is specified and not --batch-size or --max-sentences
        # Pro of --max-tokens: more flexible for MT variable sequence lengths
        # Con of --max-tokens: extra all-reduce needed to figure out optimal
        # padding without running OOM
        if expected_bsz == 0:
            expected_dim = reshaped_input_shape[0] * torch.ones(
                (1,), dtype=torch.long, device=input.device
            )
            dist.all_reduce(
                expected_dim, group=dist.group.WORLD, op=dist.ReduceOp.MAX
            )
            expected_dim = int(expected_dim.item())
            padded_input = torch.zeros(
                (expected_dim, reshaped_input_shape[1]),
                dtype=input.dtype,
                layout=input.layout,
                device=input.device,
            )
            padded_input[: reshaped_input_shape[0], :] = reshaped_input
            reshaped_input = padded_input

            padded_input_padding_mask = torch.ones(
                (expected_dim,), dtype=torch.bool, device=padded_input.device
            )
            if reshaped_input_padding_mask is not None:
                padded_input_padding_mask[: reshaped_input_shape[0]] = (
                    reshaped_input_padding_mask
                )
            else:
                padded_input_padding_mask[: reshaped_input_shape[0]] = False
            reshaped_input_padding_mask = padded_input_padding_mask

        if has_tutel:
            (
                l_aux,
                self.metadata,
                C,
                E,
                indices_,
                locations_,
                gates_,
            ) = self.gate(reshaped_input, reshaped_input_padding_mask)
            S, M = reshaped_input.size(0), reshaped_input.size(1)

            if not hasattr(self, "_tutel_dispatcher"):
                self._tutel_dispatcher = tutel_moe.fast_dispatcher(
                    E, C, M, dispatch_dtype=reshaped_input.dtype
                )
            self._tutel_dispatcher.update(
                indices_, locations_, gates_, capacity=C
            )
            dispatched_input = self._tutel_dispatcher.encode(reshaped_input)
        else:
            l_aux, combine_weights, dispatch_mask, self.metadata = self.gate(
                reshaped_input, reshaped_input_padding_mask
            )

            dispatch_mask = dispatch_mask.to(input.dtype).permute(
                1, 2, 0
            )  # S,E,C -> E,C,S
            E, C, S = dispatch_mask.size()
            M = reshaped_input.size(1)
            assert reshaped_input.size() == (S, M)
            # einsum("sec,sm->ecm")
            dispatched_input = torch.mm(
                dispatch_mask.view(E * C, S), reshaped_input
            )  # -> (E*C),M

        if self.all2all_size > 1:
            dispatched_input = self.all_to_all_wrapper(dispatched_input)

        # Re-shape after all-to-all: ecm -> gecm
        dispatched_input = dispatched_input.reshape(
            self.all2all_size, self.num_local_experts, -1, dim
        )
        chunks = dispatched_input.chunk(self.num_local_experts, dim=1)
        expert_outputs = []
        for chunk, expert in zip(chunks, self.experts):
            expert_outputs += [expert(chunk)]
        expert_output = torch.cat(expert_outputs, dim=1)

        if self.all2all_size > 1:
            expert_output = self.all_to_all_wrapper(expert_output)

        # Re-shape back: gecm -> ecm
        expert_output = expert_output.reshape(
            self.all2all_size * self.num_local_experts, -1, dim
        )

        if has_tutel:
            combined_output = self._tutel_dispatcher.decode(
                expert_output.view(E * C, M)
            )
        else:
            # einsum("sec,ecm->sm")
            combined_output = combine_weights.view(S, E * C).mm(
                expert_output.view(E * C, M)
            )

        # Remove padding here when --max-tokens is specified and not
        # --batch-size or --max-sentences
        combined_output = combined_output[: reshaped_input_shape[0], :]
        combined_output = combined_output.reshape(input.shape)
        combined_output = combined_output[: input_shape[0], :, :]

        self.record_all_to_all_stats()

        return combined_output, l_aux

    def prepare_for_inference_(self):
        self.in_generation = True

    def all_to_all_wrapper(self, input: Tensor):
        dummy_a2a = getattr(self.args, "dummy_a2a", False)
        if dummy_a2a:
            input = input.contiguous()
            output = input.detach().clone()
            return input
        # always record times, since it is not a lot of overhead
        # if we do not log it we simply clear it off in record_all_to_all_stats
        cuda_start = torch.cuda.Event(enable_timing=True)
        cuda_end = torch.cuda.Event(enable_timing=True)
        cpu_start = time.time() * 1000
        cuda_start.record()
        output = _AllToAll.apply(self.all2all_group, input)
        cuda_end.record()
        cpu_end = time.time() * 1000
        self.a2a_cpu_time_ms += cpu_end - cpu_start
        self.a2a_cuda_event_intervals.append((cuda_start, cuda_end))
        return output

    def record_all_to_all_stats(self):
        # controlled via an argument as we want to minimize any impact from
        # torch.cuda.synchronize()
        record_a2a_perf_stats = getattr(
            self.args, "record_a2a_perf_stats", False
        )
        if record_a2a_perf_stats:
            torch.cuda.synchronize()
            self.metadata["all_to_all_cpu_time_ms"] = self.a2a_cpu_time_ms
            a2a_cuda_time_ms = 0.0
            for ev_start, ev_end in self.a2a_cuda_event_intervals:
                a2a_cuda_time_ms += ev_start.elapsed_time(ev_end)
            self.metadata["all_to_all_cuda_time_ms"] = a2a_cuda_time_ms
        # reset stats
        self.a2a_cpu_time_ms = 0.0
        self.a2a_cuda_event_intervals = []



import torch
from torch import nn, Tensor
import torch.nn.functional as F
from typing import Optional, Dict
from dataclasses import dataclass
import math
from loguru import logger
from pathlib import Path
from datetime import datetime


@dataclass
class FlowTransformerConfig:
    """Configuration for Flow Transformer.

    Attributes:
        dim: Model dimension
        heads: Number of attention heads
        depth: Number of transformer layers
        seq_length: Maximum sequence length
        flow_hidden_dim: Hidden dimension for flow networks
        flow_steps: Number of flow integration steps
        dropout: Dropout rate
        attention_dropout: Dropout rate for attention
        learning_rate: Learning rate for optimizer
        weight_decay: Weight decay for optimizer
        warmup_steps: Number of warmup steps for learning rate
        max_steps: Maximum number of training steps
        batch_size: Batch size for training
        checkpoint_interval: Number of steps between checkpoints
        device: Device to use for computation
    """

    dim: int = 512
    heads: int = 8
    depth: int = 6
    seq_length: int = 1024
    flow_hidden_dim: int = 128
    flow_steps: int = 4
    dropout: float = 0.1
    attention_dropout: float = 0.1
    learning_rate: float = 1e-4
    vocab_size: int = 50000
    weight_decay: float = 0.01
    warmup_steps: int = 10000
    max_steps: int = 100000
    batch_size: int = 32
    checkpoint_interval: int = 1000
    device: torch.device = torch.device(
        "cuda" if torch.cuda.is_available() else "cpu"
    )


class FlowMLP(nn.Module):
    def __init__(self, config: FlowTransformerConfig):
        super().__init__()
        self.config = config

        # Flow network for velocity field - keep input dimension as config.dim
        self.flow_net = nn.Sequential(
            nn.Linear(config.dim, config.flow_hidden_dim),
            nn.LayerNorm(config.flow_hidden_dim),
            nn.GELU(),
            nn.Linear(config.flow_hidden_dim, config.flow_hidden_dim),
            nn.LayerNorm(config.flow_hidden_dim),
            nn.GELU(),
            nn.Linear(config.flow_hidden_dim, config.dim),
        )

        # Time embedding
        self.time_embedding = nn.Linear(1, config.dim, bias=False)
        self.time_scale = nn.Parameter(torch.ones(1))

        logger.debug(f"Initialized FlowMLP with architecture: {self.flow_net}")

    def forward(self, x: Tensor) -> Tensor:
        batch_size, seq_length, dim = x.shape

        # Initialize flow
        current_state = x

        # Integration steps
        dt = 1.0 / self.config.flow_steps
        for step in range(self.config.flow_steps):
            t = (
                torch.full(
                    (batch_size, seq_length, 1), step * dt, device=x.device
                )
                * self.time_scale
            )

            # Project time to dimension space and add to current state
            t_embedded = self.time_embedding(t)
            flow_input = current_state + t_embedded

            # Compute velocity field
            velocity = self.flow_net(flow_input)

            # Euler integration step
            current_state = current_state + velocity * dt

        return current_state


class MultiHeadAttention(nn.Module):
    def __init__(self, config: FlowTransformerConfig):
        super().__init__()
        self.config = config

        assert config.dim % config.heads == 0, "dim must be divisible by heads"
        self.head_dim = config.dim // config.heads

        self.to_qkv = nn.Linear(config.dim, 3 * config.dim, bias=False)
        self.to_out = nn.Linear(config.dim, config.dim)

        self.attention_dropout = nn.Dropout(config.attention_dropout)
        self.output_dropout = nn.Dropout(config.dropout)

        nn.init.xavier_normal_(self.to_qkv.weight, gain=0.1)
        nn.init.xavier_normal_(self.to_out.weight, gain=0.1)

    def forward(self, x: Tensor, mask: Optional[Tensor] = None) -> Tensor:
        batch_size, seq_length, _ = x.shape

        # Project to q, k, v
        qkv = self.to_qkv(x)
        q, k, v = qkv.chunk(3, dim=-1)

        # Split into heads
        q = q.view(
            batch_size, seq_length, self.config.heads, self.head_dim
        ).transpose(1, 2)
        k = k.view(
            batch_size, seq_length, self.config.heads, self.head_dim
        ).transpose(1, 2)
        v = v.view(
            batch_size, seq_length, self.config.heads, self.head_dim
        ).transpose(1, 2)

        # Compute attention scores
        scale = math.sqrt(self.head_dim)
        scores = torch.matmul(q, k.transpose(-2, -1)) / scale  # [B, H, S, S]

        if mask is not None:
            # Expand mask to match attention scores shape
            mask = mask.view(batch_size, 1, 1, seq_length)
            mask = mask.expand(-1, self.config.heads, seq_length, -1)
            scores = scores.masked_fill(~mask.bool(), float("-inf"))

        # Apply attention
        attn = F.softmax(scores, dim=-1)
        attn = self.attention_dropout(attn)

        # Compute output
        out = torch.matmul(attn, v)
        out = out.transpose(1, 2).contiguous()
        out = out.view(batch_size, seq_length, self.config.dim)

        # Project to output
        out = self.to_out(out)
        out = self.output_dropout(out)

        return out


class TransformerBlock(nn.Module):
    """Transformer block with FlowMLP."""

    def __init__(self, config: FlowTransformerConfig):
        """Initialize transformer block.

        Args:
            config: Configuration object
        """
        super().__init__()
        self.config = config

        # Layer norm for pre-attention and pre-MLP
        self.norm1 = nn.LayerNorm(config.dim)
        self.norm2 = nn.LayerNorm(config.dim)

        # Attention and MLP
        self.attention = MultiHeadAttention(config)
        self.flow_mlp = FlowMLP(config)

        # Dropout
        self.dropout = nn.Dropout(config.dropout)

    def forward(self, x: Tensor, mask: Optional[Tensor] = None) -> Tensor:
        """Process input through transformer block.

        Args:
            x: Input tensor of shape (batch_size, seq_length, dim)
            mask: Optional attention mask

        Returns:
            Output tensor of shape (batch_size, seq_length, dim)
        """
        # Attention with residual
        attn_out = self.attention(self.norm1(x), mask)
        x = x + self.dropout(attn_out)

        # FlowMLP with residual
        mlp_out = self.flow_mlp(self.norm2(x))
        x = x + self.dropout(mlp_out)

        return x


class FlowTransformer(nn.Module):
    """Transformer model with FlowMLP blocks."""

    def __init__(self, config: FlowTransformerConfig):
        """Initialize Flow Transformer.

        Args:
            config: Configuration object
        """
        super().__init__()
        self.config = config
        logger.info(f"Initializing Flow Transformer with config: {config}")

        # Token embedding and positional encoding
        self.token_embedding = nn.Embedding(config.vocab_size, config.dim)
        self.position_embedding = nn.Parameter(
            torch.zeros(1, config.seq_length, config.dim)
        )

        # Transformer blocks
        self.blocks = nn.ModuleList(
            [TransformerBlock(config) for _ in range(config.depth)]
        )

        # Final layer norm
        self.norm = nn.LayerNorm(config.dim)

        # Output projection
        self.to_logits = nn.Linear(config.dim, config.vocab_size)

        self._init_weights()
        self._setup_logging()

    def _init_weights(self):
        """Initialize model weights."""
        nn.init.normal_(self.token_embedding.weight, std=0.02)
        nn.init.normal_(self.position_embedding, std=0.02)

    def _setup_logging(self):
        """Setup logging configuration."""
        timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")
        log_path = Path(f"logs/flow_transformer_{timestamp}.log")
        log_path.parent.mkdir(parents=True, exist_ok=True)
        logger.add(log_path, rotation="500 MB")

    def forward(self, x: Tensor, mask: Optional[Tensor] = None) -> Tensor:
        """Process input through Flow Transformer.

        Args:
            x: Input tensor of shape (batch_size, seq_length)
            mask: Optional attention mask

        Returns:
            Output logits of shape (batch_size, seq_length, vocab_size)
        """
        batch_size, seq_length = x.shape

        # Embeddings
        x = self.token_embedding(x)
        x = x + self.position_embedding[:, :seq_length]

        # Process through transformer blocks
        for block in self.blocks:
            x = block(x, mask)

        # Final norm and projection
        x = self.norm(x)
        logits = self.to_logits(x)

        return logits

    def train_step(
        self, batch: Dict[str, Tensor], optimizer: torch.optim.Optimizer
    ) -> Dict[str, float]:
        """Perform one training step.

        Args:
            batch: Dictionary containing input_ids and labels
            optimizer: Optimizer instance

        Returns:
            Dictionary of metrics
        """
        optimizer.zero_grad()

        # Forward pass
        logits = self(batch["input_ids"], batch.get("attention_mask"))

        # Compute loss
        loss = F.cross_entropy(
            logits.view(-1, self.config.vocab_size),
            batch["labels"].view(-1),
            ignore_index=-100,
        )

        # Backward pass
        loss.backward()
        torch.nn.utils.clip_grad_norm_(self.parameters(), 1.0)
        optimizer.step()

        return {"loss": loss.item()}

    def save_checkpoint(self, path: Path) -> None:
        """Save model checkpoint.

        Args:
            path: Path to save checkpoint
        """
        checkpoint = {
            "model_state_dict": self.state_dict(),
            "config": self.config,
            "timestamp": datetime.now().isoformat(),
        }
        path.parent.mkdir(parents=True, exist_ok=True)
        torch.save(checkpoint, path)
        logger.info(f"Checkpoint saved to {path}")

    @classmethod
    def load_checkpoint(cls, path: Path) -> "FlowTransformer":
        """Load model from checkpoint.

        Args:
            path: Path to checkpoint file

        Returns:
            Loaded FlowTransformer model
        """
        checkpoint = torch.load(path)
        model = cls(checkpoint["config"])
        model.load_state_dict(checkpoint["model_state_dict"])
        logger.info(
            f"Model loaded from {path} (saved at {checkpoint['timestamp']})"
        )
        return model


def create_optimizer(
    model: FlowTransformer, config: FlowTransformerConfig
) -> torch.optim.AdamW:
    """Create optimizer with weight decay fix.

    Args:
        model: FlowTransformer model
        config: Configuration object

    Returns:
        AdamW optimizer
    """
    # Separate parameters that should and shouldn't use weight decay
    decay = set()
    no_decay = set()

    for mn, m in model.named_modules():
        for pn, p in m.named_parameters():
            fpn = f"{mn}.{pn}" if mn else pn
            if pn.endswith("bias") or "layer_norm" in fpn:
                no_decay.add(fpn)
            else:
                decay.add(fpn)

    param_dict = {pn: p for pn, p in model.named_parameters()}
    optim_groups = [
        {
            "params": [param_dict[pn] for pn in sorted(list(decay))],
            "weight_decay": config.weight_decay,
        },
        {
            "params": [param_dict[pn] for pn in sorted(list(no_decay))],
            "weight_decay": 0.0,
        },
    ]

    return torch.optim.AdamW(optim_groups, lr=config.learning_rate)


# Create configuration
config = FlowTransformerConfig(
    dim=512,  # Model dimension
    heads=8,  # Number of attention heads
    depth=6,  # Number of transformer layers
    seq_length=1024,  # Maximum sequence length
    vocab_size=50000,  # Vocabulary size
    flow_steps=4,  # Number of flow integration steps
)

# Initialize model
model = FlowTransformer(config)
model = model.to(config.device)

# Create a sample input
batch_size = 4
seq_length = 512
input_ids = torch.randint(
    0, config.vocab_size, (batch_size, seq_length), device=config.device
)

# Optional attention mask (1 for tokens to attend to, 0 for tokens to ignore)
attention_mask = torch.ones((batch_size, seq_length), device=config.device)

# Forward pass
with torch.no_grad():
    output_logits = model(input_ids, attention_mask)

print(f"Input shape: {input_ids.shape}")
print(
    f"Output shape: {output_logits.shape}"
)  # Should be [batch_size, seq_length, vocab_size]

# Get predictions
predictions = torch.argmax(output_logits, dim=-1)
print(
    f"Predictions shape: {predictions.shape}"
)  # Should be [batch_size, seq_length]



import torch


def kv_cache_with_update(K, V, qt, kt, vt):
    """
    Single-head KV cache update with Dynamic Memory Compression (DMC).

    Parameters:
        K (torch.Tensor): The key matrix (batch, seqlen, dimension).
        V (torch.Tensor): The value matrix (batch, seqlen, dimension).
        qt (torch.Tensor): The current query vector (batch, seqlen, dimension).
        kt (torch.Tensor): The current key vector (batch, seqlen, dimension).
        vt (torch.Tensor): The current value vector (batch, seqlen, dimension).

    Returns:
        tuple: Updated K, V, qt, kt tensors.

    Example:
    """
    # Calculate alpha_t and omega_t using the first element of kt and qt respectively
    # Assume we use the first element of the last dimension for decision and weighting
    alpha_t = torch.round(torch.sigmoid(kt[:, :, 0]))  # Shape (batch, seqlen)
    omega_t = torch.sigmoid(qt[:, :, 0])  # Shape (batch, seqlen)

    # Extend alpha_t and omega_t for element-wise operations
    alpha_t = alpha_t.unsqueeze(-1)  # Shape (batch, seqlen, 1)
    omega_t = omega_t.unsqueeze(-1)  # Shape (batch, seqlen, 1)

    # Initialize z_t if not provided, we'll assume it starts with the initial omega_t values
    zt = omega_t.clone()

    # ACCUMULATE
    # Update keys and values with weighted average only where alpha_t is 1
    accumulate_mask = alpha_t == 1
    K_new = (K * zt + kt * omega_t) / (zt + omega_t)
    V_new = (V * zt + vt * omega_t) / (zt + omega_t)

    # Only update where accumulate condition is met
    K = torch.where(accumulate_mask, K_new, K)
    V = torch.where(accumulate_mask, V_new, V)

    # APPEND
    # Only update where accumulate condition is not met
    append_mask = alpha_t != 1
    K = torch.where(append_mask, kt, K)
    V = torch.where(append_mask, vt, V)

    # Update z_t considering whether to accumulate or just set to omega_t
    zt = torch.where(accumulate_mask, zt + omega_t, omega_t)

    # Reset the first elements used in kt and qt to 0
    kt[:, :, 0] = 0
    qt[:, :, 0] = 0

    return K, V, qt, kt


# # Example of usage:
# batch_size = 2
# seqlen = 5
# dim = 3

# K = torch.randn(batch_size, seqlen, dim)  # Key matrix
# V = torch.randn(batch_size, seqlen, dim)  # Value matrix
# qt = torch.randn(batch_size, seqlen, dim)  # Query vectors
# kt = torch.randn(batch_size, seqlen, dim)  # Key vectors
# vt = torch.randn(batch_size, seqlen, dim)  # Value vectors

# K_updated, V_updated, qt_updated, kt_updated = kv_cache_with_update(
#     K, V, qt, kt, vt
# )
# print("Updated K:", K_updated)
# print("Updated V:", V_updated)



import math

from einops.layers.torch import Rearrange, Reduce
from torch import nn


def make_layer(inplanes, planes, block, n_blocks, stride=1):
    downsample = None
    if stride != 1 or inplanes != planes * block.expansion:
        # output size won't match input so adjust residual
        downsample = nn.Sequential(
            nn.Conv2d(
                inplanes,
                planes * block.expansion,
                kernel_size=1,
                stride=stride,
                bias=False,
            ),
        )
        nn.BatchNorm2d(planes * block.expansion),

    return nn.Sequential(
        block(inplanes, planes, stride, downsample),
        *[block(planes * block.expansion, planes) for _ in range(1, n_blocks)],
    )


class ResNet(nn.Module):
    """
    Resnet implementation.

    Usage:
        from zeta.nn import ResNet

        x = torch.randn(1, 3, 224, 224)
        net = ResNet()
        net(x)



    """

    def __init__(self, block, layers, num_classes=1000):
        super().__init__()

        e = block.expansion

        self.resnet = nn.Sequential(
            Rearrange("b c h w -> b c h w", c=3, h=224, w=224),
            nn.Conv2d(3, 64, kernel_size=7, stride=2, padding=3, bias=False),
            nn.BatchNorm2d(64),
            nn.ReLU(inplace=True),
            nn.MaxPool2d(kernel_size=3, stride=2, padding=1),
            make_layer(64, 64, block, layers[0], stride=1),
            make_layer(64 * e, 128, block, layers[1], stride=2),
            make_layer(128 * e, 256, block, layers[2], stride=2),
            make_layer(256 * e, 512, block, layers[3], stride=2),
            # combind avg pool and view in one averaging operation
            Reduce("b c h w -> b c", "mean"),
            nn.Linear(512 * e, num_classes),
        )

    def __call__(self):
        """
        Call the forward pass.
        """
        for m in self.resnet.modules():
            if isinstance(m, nn.Conv2d):
                n = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
                m.weight.data.normal_(0, math.sqrt(2.0 / n))
            elif isinstance(m, nn.BatchNorm2d):
                m.weight.data.fill_(1)
                m.bias.data.zero_()
        return self.resent



import torch


class GradientAscent:
    """
    Gradient Ascent Optimizer

    Optimizer that performs gradient ascent on the parameters of the model.

    Args:
        parameters (iterable): iterable of parameters to optimize or dicts defining parameter groups
        lr (float, optional): learning rate (default: 0.01)
        momentum (float, optional): momentum factor (default: 0.9)
        beta (float, optional): beta factor (default: 0.999)
        eps (float, optional): epsilon (default: 1e-8)
        nesterov (bool, optional): enables Nesterov accelerated gradient (default: False)
        clip_value (float, optional): gradient clipping value (default: None)
        lr_decay (float, optional): learning rate decay (default: None)
        warmup_steps (int, optional): warmup steps (default: 0)
        logging_interval (int, optional): logging interval (default: 10)


    Attributes:
        defaults (dict): default optimization options
        parameters (iterable): iterable of parameters to optimize or dicts defining parameter groups
        lr (float): learning rate
        momentum (float): momentum factor
        beta (float): beta factor
        eps (float): epsilon
        v (dict): momentum
        m (dict): adaptive learning rate

    Example:
        >>> optimizer = GradientAscent(model.parameters(), lr=0.01)
        >>> optimizer.zero_grad()
        >>> loss_fn(model(input), target).backward()
        >>> optimizer.step()


    """

    def __init__(
        self,
        parameters,
        lr=0.01,
        momentum=0.9,
        beta=0.999,
        eps=1e-8,
        nesterov=False,
        clip_value=None,
        lr_decay=None,
        warmup_steps=0,
        logging_interval=10,
    ):
        self.parameters = list(parameters)
        self.lr = lr
        self.momentum = momentum
        self.beta = beta
        self.eps = eps
        # Nesterov accelerated gradient NAG => Provides a lookahead in the direction of the parameter updates => optimizer converge faster
        self.nesterov = nesterov
        # Gradient Clipping => Prevents exploding gradients
        self.clip_value = clip_value
        # Learning Rate Decay => Prevents oscillations
        self.lr_decay = lr_decay
        self.warmup_steps = warmup_steps
        self.logging_interval = logging_interval

        self.step_count = 0

        # Initalize momentum and adaptive learning rate
        self.v = {p: torch.zeros_like(p.data) for p in self.parameters}
        self.m = {p: torch.zeros_like(p.data) for p in self.parameters}

    def step(self):
        self.step_count += 1
        """Step function for gradient ascent optimizer"""
        for param in self.parameters:
            try:
                if param.grad is not None:
                    if self.clip_value:
                        torch.nn.utils.clip_grad_value_(
                            param.grad, self.clip_value
                        )

                    # Nesterov Accelerated Gradient
                    if self.nesterov:
                        grad = param.grad + self.momentum * self.v[param]
                    else:
                        grad = param.grad

                    # Momentum
                    self.v[param] = self.momentum * self.v[param] + grad

                    # Adaptive learning rate
                    self.m[param] = (
                        self.beta * self.m[param] + (1 - self.beta) * grad**2
                    )
                    adapted_lr = self.lr / (
                        torch.sqrt(self.m[param]) + self.eps
                    )

                    # Warmup Learning Rate
                    if self.step_count <= self.warmup_steps:
                        warmup_factor = self.step_count / float(
                            self.warmup_steps
                        )
                        adapted_lr *= warmup_factor

                    # Gradient Ascent
                    param.data.add_(adapted_lr * self.v[param])

                    # Learning Rate Decay
                    if self.lr_decay:
                        self.lr *= self.lr_decay

                if self.step_count % self.logging_interval == 0:
                    print(
                        f"Step: {self.step_count}, Learning Rate: {self.lr},"
                        f" Gradient Norm: {torch.norm(param.grad)}"
                    )

            except Exception as error:
                print(f"Exception during optimization: {error}")

    def zero_grad(self):
        """Zero the gradient of the parameters"""
        for param in self.parameters:
            if param.grad is not None:
                param.grad.detach_()
                param.grad.zero_()



import pickle
from pathlib import Path

import torch
from beartype import beartype
from beartype.typing import Callable, Optional
from torch.nn import Module


# helpers
def exists(v):
    return v is not None


@beartype
def save_load(
    save_method_name="save",
    load_method_name="load",
    config_instance_var_name="_config",
    init_and_load_classmethod_name="init_and_load",
    version: Optional[str] = None,
    pre_save_hook: Optional[Callable[[Module], None]] = None,
    post_load_hook: Optional[Callable[[Module], None]] = None,
    compress: Optional[bool] = False,
    partial_load: Optional[bool] = False,
    *args,
    **kwargs,
):
    """Base decorator for save and load methods for torch.nn.Module subclasses.

    Args:
        save_method_name (str, optional): _description_. Defaults to "save".
        load_method_name (str, optional): _description_. Defaults to "load".
        config_instance_var_name (str, optional): _description_. Defaults to "_config".
        init_and_load_classmethod_name (str, optional): _description_. Defaults to "init_and_load".
        version (Optional[str], optional): _description_. Defaults to None.
        pre_save_hook (Optional[Callable[[Module], None]], optional): _description_. Defaults to None.
        post_load_hook (Optional[Callable[[Module], None]], optional): _description_. Defaults to None.
        compress (Optional[bool], optional): _description_. Defaults to False.
        partial_load (Optional[bool], optional): _description_. Defaults to False.
    """

    def _save_load(klass):
        assert issubclass(
            klass, Module
        ), "save_load should decorate a subclass of torch.nn.Module"

        _orig_init = klass.__init__

        def __init__(self, *args, **kwargs):
            _config = pickle.dumps((args, kwargs))
            setattr(self, config_instance_var_name, _config)
            _orig_init(self, *args, **kwargs)

        def _save(self, path, overwrite=True):
            if pre_save_hook:
                pre_save_hook(self)

            path = Path(path)
            assert overwrite or not path.exists()
            pkg = dict(
                model=self.state_dict(),
                config=getattr(self, config_instance_var_name),
                version=version,
            )
            torch.save(pkg, str(path), _use_new_zipfile_serialization=compress)

        def _load(self, path, strict=True):
            path = Path(path)
            assert path.exists()
            pkg = torch.load(str(path), map_location="cpu", weights_only=True)

            if (
                exists(version)
                and exists(pkg["version"])
                and version.parse(version) != version.parse(pkg["version"])
            ):
                self.print(f'loading saved model at version {pkg["version"]},')

            model_dict = self.state_dict()
            if partial_load:
                model_dict.update(pkg["model"])
                self.load_state_dict(model_dict, strict=strict)
            else:
                self.load_state_dict(pkg["model"], strict=strict)

            if post_load_hook:
                post_load_hook(self)

        @classmethod
        def _init_and_load_from(cls, path, strict=True):
            path = Path(path)
            assert path.exists()
            pkg = torch.load(str(path), map_location="cpu", weights_only=True)
            assert (
                "config" in pkg
            ), "model configs were not found in this saved checkpoint"

            config = pickle.loads(pkg["config"])
            args, kwargs = config
            model = cls(*args, **kwargs)

            _load(model, path, strict=strict)
            return model

        klass.__init__ = __init__
        setattr(klass, save_method_name, _save)
        setattr(klass, load_method_name, _load)
        setattr(klass, init_and_load_classmethod_name, _init_and_load_from)

        return klass

    return _save_load



import torch
import triton

from typing import Callable
from zeta.experimental.triton.activations.functions import Functions

BLOCK_SIZE = 1024


def apply_activation(
    x: torch.Tensor, activation_fn: Callable[..., torch.Tensor], *args, **kwargs
):
    if not x.is_cuda:
        raise ValueError("Input tensor must be on CUDA.")

    output = torch.empty_like(x)
    n_elements = output.numel()

    def grid(meta):
        return (triton.cdiv(n_elements, meta["BLOCK_SIZE"]),)

    activation_args = [x, output] + list(args)

    if "n_elements" not in kwargs:
        kwargs["n_elements"] = n_elements

    activation_fn[grid](*activation_args, BLOCK_SIZE=1024, **kwargs)

    return output


def tanh_activation(x: torch.Tensor):
    return apply_activation(x, Functions.tanh_activation_kernel)


def hard_tanh_activation(x: torch.Tensor):
    return apply_activation(x, Functions.hard_tanh_activation_kernel)


def relu_activation(x: torch.Tensor):
    return apply_activation(x, Functions.relu_activation_kernel)


def relu6_activation(x: torch.Tensor):
    return apply_activation(x, Functions.relu6_activation_kernel)


def leaky_relu_activation(x: torch.Tensor, alpha: float = 0.2):
    return apply_activation(
        x, Functions.leaky_relu_activation_kernel, alpha=alpha
    )


def smooth_relu_activation(x: torch.Tensor, beta: float = 2.0):
    # Make input tensor contiguous if needed
    if not x.is_contiguous():
        x = x.contiguous()

    return apply_activation(
        x, Functions.smooth_relu_activation_kernel, beta=beta
    )


def softsign_activation(x: torch.Tensor):
    return apply_activation(x, Functions.softsign_activation_kernel)


def softplus_activation(x: torch.Tensor):
    return apply_activation(x, Functions.softplus_activation_kernel)


def sigmoid_activation(x: torch.Tensor):
    return apply_activation(x, Functions.sigmoid_activation_kernel)


def hard_sigmoid_activation(x: torch.Tensor):
    return apply_activation(x, Functions.hard_sigmoid_activation_kernel)


def silu_activation(x: torch.Tensor):
    return apply_activation(x, Functions.silu_activation_kernel)


def hard_silu_activation(x: torch.Tensor):
    return apply_activation(x, Functions.hard_silu_activation_kernel)


def softmax_activation(x: torch.Tensor):
    return apply_activation(x, Functions.softmax_activation_kernel)


def gelu_activation(x: torch.Tensor, approximate: bool = True):
    return apply_activation(x, Functions.gelu_activation_kernel, approximate)


def swiglu_activation(x: torch.Tensor):
    return apply_activation(x, Functions.swiglu_activation_kernel)



"""Neural network modules. zeta/nn"""

from zeta.nn.attention import *  # noqa: F403
from zeta.nn.biases import *  # noqa: F403
from zeta.nn.embeddings import *  # noqa: F403
from zeta.nn.modules import *  # noqa: F403



import torch
from torch import nn


class LeakyRELU(nn.Module):
    """LeakyReLU activation function.

    Args:
        nn (_type_): _description_

    Returns:
        _type_: _description_
    """

    __constants__ = ["inplace", "negative_slope"]
    inplace: bool
    negative_sloop: float

    def __init__(
        self,
        negative_slope: float = 1e-2,
        inplace: bool = False,
    ) -> None:
        super().__init__()
        self.negative_slope = negative_slope
        self.inplace = inplace

    def forward(
        self,
        input: torch.Tensor,
    ) -> torch.Tensor:
        """Forward pass of the LeakyReLU module.

        Args:
            input (torch.Tensor): _description_

        Returns:
            torch.Tensor: _description_
        """
        return torch.where(input >= 0.0, input, input * self.negative_slope)

    def extra_repr(self) -> str:
        """Extra information about this module.

        Returns:
            str: _description_
        """
        inplace_str = ", inplace=True" if self.inplace else ""
        return f"negative_slope={self.negative_slope}{inplace_str}"



import random
from collections import deque

import numpy as np
import torch


class HindsightExperienceReplay:
    """
    Hindsight experience replay buffer.

    Parameters
    ----------
    state_dim : int
        the dimension of the state
    action_dim : int
        the dimension of the action
    buffer_size : int
        the maximum size of the buffer
    batch_size : int
        the size of the mini-batch
    goal_sampling_strategy : function
        the goal sampling strategy to use

    Example:
    import torch
    from hindsight import HindsightExperienceReplay
    from numpy import np





    # Define a goal sampling strategy
    def goal_sampling_strategy(goals):
        noise = torch.randn_like(goals) * 0.1
        return goals + noise


    # Define the dimensions of the state and action spaces, the buffer size, and the batch size
    state_dim = 10
    action_dim = 2
    buffer_size = 10000
    batch_size = 64

    # Create an instance of the HindsightExperienceReplay class
    her = HindsightExperienceReplay(
        state_dim, action_dim, buffer_size, batch_size, goal_sampling_strategy
    )

    # Store a transition
    state = np.random.rand(state_dim)
    action = np.random.rand(action_dim)
    reward = np.random.rand()
    next_state = np.random.rand(state_dim)
    done = False
    goal = np.random.rand(state_dim)
    her.store_transition(state, action, reward, next_state, done, goal)

    # Sample a mini-batch of transitions
    sampled_transitions = her.sample()
    if sampled_transitions is not None:
        states, actions, rewards, next_states, dones, goals = sampled_transitions



    """

    def __init__(
        self,
        state_dim,
        action_dim,
        buffer_size,
        batch_size,
        goal_sampling_strategy,
    ):
        self.state_dim = state_dim
        self.action_dim = action_dim
        self.buffer_size = buffer_size
        self.batch_size = batch_size
        self.buffer = deque(maxlen=buffer_size)
        self.goal_sampling_strategy = goal_sampling_strategy

    def store_transition(self, state, action, reward, next_state, done, goal):
        """Store and transitions"""
        transition = (state, action, reward, next_state, done, goal)
        self.buffer.append(transition)

        # Store additional transition where the goal is replaced with the achieved state
        achieved_goal = next_state
        transition = (state, action, reward, next_state, done, achieved_goal)
        self.buffer.append(transition)

    def sample(self):
        """Sample a mini-batch of transitions"""
        if len(self.buffer) < self.batch_size:
            return None

        mini_batch = random.sample(self.buffer, self.batch_size)

        states, actions, rewards, next_states, dones, goals = zip(*mini_batch)

        states = torch.FloatTensor(states)
        actions = torch.FloatTensor(actions)
        rewards = torch.FloatTensor(rewards).unsqueeze(1)
        next_states = torch.FloatTensor(next_states)
        dones = torch.FloatTensor(np.float32(dones)).unsqueeze(1)
        goals = torch.FloatTensor(goals)

        # Apply goal sampling strategy
        goals = self.goal_sampling_strategy(goals)

        return states, actions, rewards, next_states, dones, goals

    def __len__(self):
        """Return the length of the buffer"""
        return len(self.buffer)



from torch import Tensor, nn

from zeta.nn.quant.bitlinear import absmax_quantize


class QuantizedLN(nn.Module):
    def __init__(
        self,
        normalized_shape,
        bits: int = 8,
        eps=1e-5,
        element_wise_affine=True,
    ):
        """
        Initializes a QuantizedLN module.

        Args:
            normalized_shape (int or tuple): The expected input shape.
            bits (int, optional): Number of bits for quantization. Defaults to 8.
            eps (float, optional): A value added to the denominator for numerical stability. Defaults to 1e-5.
            element_wise_affine (bool, optional): Whether to include learnable affine parameters. Defaults to True.

        Examples::
        x = torch.randn(128, 10)
        ln = QuantizedLN(10)
        output = ln(x)
        print(output)

        """
        super().__init__()
        self.bits = bits
        self.ln = nn.LayerNorm(
            normalized_shape, eps=eps, elementwise_affine=element_wise_affine
        )

    def forward(self, x: Tensor):
        """
        Forward pass of the QuantizedLN module.

        Args:
            x (torch.Tensor): Input tensor.

        Returns:
            torch.Tensor: Output tensor after applying quantization and layer normalization.
        """
        _, x_dequant = absmax_quantize(x, bits=self.bits)
        return self.ln(x_dequant)



import torch
from einops import rearrange


def yolo(input, num_classes, num_anchors, anchors, stride_h, stride_w):
    """
    Yolo for object detection

    Args:
        input: input tensor
        num_classes: number of classes
        num_anchors: number of anchors
        anchors: anchor boxes
        stride_h: stride in height
        stride_w: stride in width

    Returns:
        predicted_bboxes: predicted bounding boxes

    Usage:
    >>> import torch
    >>> from zeta.nn.modules.yolo import yolo
    >>> input = torch.randn(1, 255, 13, 13)
    >>> num_classes = 80
    >>> num_anchors = 3
    >>> anchors = [
    ...     [10, 13],
    ...     [16, 30],
    ...     [33, 23],
    ...     [30, 61],
    ...     [62, 45],
    ...     [59, 119],
    ...     [116, 90],
    ...     [156, 198],
    ...     [373, 326],
    ... ]
    >>> stride_h = 32
    >>> stride_w = 32
    >>> predicted_bboxes = yolo(input, num_classes, num_anchors, anchors, stride_h, stride_w)
    >>> print(predicted_bboxes.shape)
    torch.Size([1, 507, 85])

    """
    raw_predictions = rearrange(
        input,
        "b (anchor prediction) h w -> prediction b anchor h w",
        anchor=num_anchors,
        prediction=5 + num_classes,
    )
    anchors = torch.FloatTensor(anchors).to(input.device)
    anchor_sizes = rearrange(anchors, "anchor dim -> dim () anchor () ()")

    _, _, _, in_h, in_w = raw_predictions.shape
    grid_h = rearrange(torch.arange(in_h).float(), "h -> () () h ()").to(
        input.device
    )
    grid_w = rearrange(torch.arange(in_w).float(), "w -> () () () w").to(
        input.device
    )

    predicted_bboxes = torch.zeros_like(raw_predictions)
    predicted_bboxes[0] = (
        raw_predictions[0].sigmoid() + grid_w
    ) * stride_w  # center x
    predicted_bboxes[1] = (
        raw_predictions[1].sigmoid() + grid_h
    ) * stride_h  # center y
    predicted_bboxes[2:4] = (
        raw_predictions[2:4].exp() * anchor_sizes
    )  # bbox width and height
    predicted_bboxes[4] = raw_predictions[4].sigmoid()  # confidence
    predicted_bboxes[5:] = raw_predictions[5:].sigmoid()  # class predictions
    # merging all predicted bboxes for each image
    return rearrange(
        predicted_bboxes, "prediction b anchor h w -> b (anchor h w) prediction"
    )



from typing import List

import torch
from torch import Tensor
from torch.optim.optimizer import Optimizer


class SophiaG(Optimizer):
    """
    SophiaG optimizer class is an optimizer that uses hessians to
    adaptively update the learning rate per parameter.

    Args
        params (iterable): iterable of parameters to optimize or dicts defining
            parameter groups
        lr (float, optional): learning rate (default: 1e-4)
        betas (Tuple[float, float], optional): coefficients used for computing
            running averages of gradient and its square (default: (0.9, 0.99))
        rho (float, optional): rho parameter (default: 0.04)
        weight_decay (float, optional): weight decay (L2 penalty) (default: 0.0)
        maximize (bool, optional): maximize the loss function (default: False)
        capturable (bool, optional): capture the batch size (default: False)
        dynamic (bool, optional): use dynamic learning rate (default: False)

    Example:
    >>> import torch
    >>> from zeta.optim import SophiaG
    >>> from zeta.training import get_lr_scheduler_with_warmup
    >>> from zeta.training import Accelerator
    >>> from torch.utils.data import DataLoader
    >>> from transformers import AdamW
    >>> from transformers import AutoTokenizer
    >>> from transformers import AutoModelForSequenceClassification
    >>> from transformers import default_data_collator
    >>> from transformers import TrainingArguments
    >>> from transformers import Trainer
    >>> from datasets import load_dataset
    >>> tokenizer = AutoTokenizer.from_pretrained("bert-base-cased")
    >>> model = AutoModelForSequenceClassification.from_pretrained("bert-base-cased", num_labels=2)
    >>> dataset = load_dataset("glue", "mrpc")
    >>> dataset = dataset.remove_columns(["idx", "label"])
    >>> dataset = dataset.rename_column("sentence1", "premise")
    >>> dataset = dataset.rename_column("sentence2", "hypothesis")
    >>> dataset = dataset.map(lambda x: tokenizer(x["premise"], x["hypothesis"], truncation=True, padding="max_length"), batched=True)
    >>> dataset = dataset.map(lambda x: {"labels": x["label"]}, batched=True)
    >>> dataset.set_format("torch", columns=["input_ids", "attention_mask", "labels"])
    >>> train_dataset, eval_dataset = dataset["train"], dataset["validation"]
    >>> data_collator = default_data_collator
    >>> training_args = TrainingArguments(
    ...     output_dir="./results",
    ...     num_train_epochs=3,
    ...     per_device_train_batch_size=32,
    ...     per_device_eval_batch_size=32,
    ...     warmup_steps=500,
    ...     weight_decay=0.01,
    ...     logging_dir="./logs",
    ...     logging_steps=10,
    ...     evaluation_strategy="epoch",
    ...     save_strategy="epoch",
    ...     load_best_model_at_end=True,
    ...     metric_for_best_model="accuracy",
    ...     greater_is_better=True,
    ... )
    >>> optimizer = SophiaG(model.parameters(), lr=5e-5, betas=(0.9, 0.999), weight_decay=0.01)
    >>> lr_scheduler = get_lr_scheduler_with_warmup(optimizer, "linear", num_warmup_steps=500, max_train_steps=training_args.max_steps)
    >>> accelerator = Accelerator()
    >>> trainer = Trainer(
    ...     model=model,
    ...     args=training_args,
    ...     train_dataset=train_dataset,
    ...     eval_dataset=eval_dataset,
    ...     data_collator=data_collator,
    ...     tokenizer=tokenizer,
    ...     optimizers=(optimizer, None),
    ...     lr_schedulers=(lr_scheduler, None),
    ...     compute_metrics=lambda p: {"accuracy": p["eval_accuracy"]},
    ... )
    >>> trainer.train()


    """

    def __init__(
        self,
        params,
        lr=1e-4,
        betas=(0.965, 0.99),
        rho=0.04,
        weight_decay=1e-1,
        *,
        maximize: bool = False,
        capturable: bool = False,
        dynamic: bool = False,
    ):
        """
        Initialize the optimizer.
        """
        if not 0.0 <= lr:
            raise ValueError(f"Invalid learning rate: {lr}")
        if not 0.0 <= betas[0] < 1.0:
            raise ValueError(f"Invalid beta parameter at index 0: {betas[0]}")
        if not 0.0 <= betas[1] < 1.0:
            raise ValueError(f"Invalid beta parameter at index 1: {betas[1]}")
        if not 0.0 <= rho:
            raise ValueError(f"Invalid rho parameter at index 1: {rho}")
        if not 0.0 <= weight_decay:
            raise ValueError(f"Invalid weight_decay value: {weight_decay}")
        defaults = dict(
            lr=lr,
            betas=betas,
            rho=rho,
            weight_decay=weight_decay,
            maximize=maximize,
            capturable=capturable,
            dynamic=dynamic,
        )
        super().__init__(params, defaults)

    def __setstate__(self, state):
        """
        Set the state of the optimizer.
        """
        super().__setstate__(state)
        for group in self.param_groups:
            group.setdefault("maximize", False)
            group.setdefault("capturable", False)
            group.setdefault("dynamic", False)
        state_values = list(self.state.values())
        step_is_tensor = (len(state_values) != 0) and torch.is_tensor(
            state_values[0]["step"]
        )
        if not step_is_tensor:
            for s in state_values:
                s["step"] = torch.tensor(float(s["step"]))

    @torch.no_grad()
    def update_hessian(self):
        """
        Update the hessian.
        """
        for group in self.param_groups:
            beta1, beta2 = group["betas"]
            for p in group["params"]:
                if p.grad is None:
                    continue
                state = self.state[p]

                if len(state) == 0:
                    state["step"] = (
                        torch.zeros((1,), dtype=torch.float, device=p.device)
                        if self.defaults["capturable"]
                        else torch.tensor(0.0)
                    )
                    state["exp_avg"] = torch.zeros_like(
                        p, memory_format=torch.preserve_format
                    )
                    state["hessian"] = torch.zeros_like(
                        p, memory_format=torch.preserve_format
                    )

                if "hessian" not in state.keys():
                    state["hessian"] = torch.zeros_like(
                        p, memory_format=torch.preserve_format
                    )

                state["hessian"].mul_(beta2).addcmul_(
                    p.grad, p.grad, value=1 - beta2
                )

    @torch.no_grad()
    def update_exp_avg(self):
        """
        Update the exponential average.
        """
        for group in self.param_groups:
            beta1, beta2 = group["betas"]
            for p in group["params"]:
                if p.grad is None:
                    continue
                state = self.state[p]
                state["exp_avg"].mul_(beta1).add_(p.grad, alpha=1 - beta1)

    @torch.no_grad()
    def step(self, closure=None, bs=5120):
        """
        Perform a step of the optimizer.
        """
        loss = None
        if closure is not None:
            with torch.enable_grad():
                loss = closure()

        self.update_hessian()
        self.update_exp_avg()

        for group in self.param_groups:
            params_with_grad = []
            grads = []
            exp_avgs = []
            state_steps = []
            hessian = []
            beta1, beta2 = group["betas"]

            for p in group["params"]:
                if p.grad is None:
                    continue
                params_with_grad.append(p)

                if p.grad.is_sparse:
                    raise RuntimeError("Hero does not support sparse gradients")
                grads.append(p.grad)
                state = self.state[p]
                # State initialization
                if len(state) == 0:
                    state["step"] = (
                        torch.zeros((1,), dtype=torch.float, device=p.device)
                        if self.defaults["capturable"]
                        else torch.tensor(0.0)
                    )
                    state["exp_avg"] = torch.zeros_like(
                        p, memory_format=torch.preserve_format
                    )
                    state["hessian"] = torch.zeros_like(
                        p, memory_format=torch.preserve_format
                    )

                if "hessian" not in state.keys():
                    state["hessian"] = torch.zeros_like(
                        p, memory_format=torch.preserve_format
                    )

                exp_avgs.append(state["exp_avg"])
                state_steps.append(state["step"])
                hessian.append(state["hessian"])

                if self.defaults["capturable"]:
                    bs = (
                        torch.ones((1,), dtype=torch.float, device=p.device)
                        * bs
                    )

            self._sophiag(
                params_with_grad,
                grads,
                exp_avgs,
                hessian,
                state_steps,
                bs=bs,
                beta1=beta1,
                beta2=beta2,
                rho=group["rho"],
                lr=group["lr"],
                weight_decay=group["weight_decay"],
                maximize=group["maximize"],
                capturable=group["capturable"],
            )

        return loss

    def _sophiag(
        self,
        params: List[Tensor],
        grads: List[Tensor],
        exp_avgs: List[Tensor],
        hessian: List[Tensor],
        state_steps: List[Tensor],
        capturable: bool = False,
        *,
        bs: int,
        beta1: float,
        beta2: float,
        rho: float,
        lr: float,
        weight_decay: float,
        maximize: bool,
    ):
        """
        SophiaG function.
        """
        if not all(isinstance(t, torch.Tensor) for t in state_steps):
            raise RuntimeError(
                "API has changed, `state_steps` argument must contain a list of"
                " singleton tensors"
            )

        self._single_tensor_sophiag(
            params,
            grads,
            exp_avgs,
            hessian,
            state_steps,
            bs=bs,
            beta1=beta1,
            beta2=beta2,
            rho=rho,
            lr=lr,
            weight_decay=weight_decay,
            maximize=maximize,
            capturable=capturable,
        )

    def _single_tensor_sophiag(
        self,
        params: List[Tensor],
        grads: List[Tensor],
        exp_avgs: List[Tensor],
        hessian: List[Tensor],
        state_steps: List[Tensor],
        *,
        bs: int,
        beta1: float,
        beta2: float,
        rho: float,
        lr: float,
        weight_decay: float,
        maximize: bool,
        capturable: bool,
    ):
        """
        SophiaG function for single tensor.
        """
        for i, param in enumerate(params):
            grad = grads[i] if not maximize else -grads[i]
            exp_avg = exp_avgs[i]
            hess = hessian[i]
            step_t = state_steps[i]

            if capturable:
                assert param.is_cuda
                assert step_t.is_cuda
                assert bs.is_cuda

            if torch.is_complex(param):
                grad = torch.view_as_real(grad)
                exp_avg = torch.view_as_real(exp_avg)
                hess = torch.view_as_real(hess)
                param = torch.view_as_real(param)

            # update step
            step_t += 1

            # Perform stepweight decay
            param.mul_(1 - lr * weight_decay)

            # Decay the first and second moment running average coefficient
            exp_avg.mul_(beta1).add_(grad, alpha=1 - beta1)

            if capturable:
                step_size = lr
                step_size_neg = step_size.neg()

                ratio = (exp_avg.abs() / (rho * bs * hess + 1e-15)).clamp(
                    None, 1
                )
                param.addcmul_(exp_avg.sign(), ratio, value=step_size_neg)
            else:
                step_t.item()
                step_size_neg = -lr

                ratio = (exp_avg.abs() / (rho * bs * hess + 1e-15)).clamp(
                    None, 1
                )
                param.addcmul_(exp_avg.sign(), ratio, value=step_size_neg)



import torch


def pairwise_distances(x):
    # Compute pairwise distance matrix
    diff = x.unsqueeze(1) - x.unsqueeze(0)
    return torch.sqrt((diff**2).sum(2))


def function_for_x(
    xi, sigma_i, N, J, alpha, beta, gamma, epsilon_a, epsilon_r, R, D
):
    dists = pairwise_distances(xi)
    mask = (dists < R).float() - torch.eye(N)

    interaction_term = mask.unsqueeze(2) * (
        sigma_i.unsqueeze(0) - sigma_i.unsqueeze(1)
    )
    interaction_sum = interaction_term.sum(1)

    # Define dynamics for x based on our assumptions
    dx = J * interaction_sum + alpha * xi - beta * (xi**3)
    return dx


def function_for_sigma(
    xi, sigma_i, N, J, alpha, beta, gamma, epsilon_a, epsilon_r, R, D
):
    dists = pairwise_distances(xi)
    mask = (dists < R).float() - torch.eye(N)

    interaction_term = mask.unsqueeze(2) * (xi.unsqueeze(0) - xi.unsqueeze(1))
    interaction_sum = interaction_term.sum(1)

    # Define dynamics for sigma based on our assumptions
    d_sigma = (
        gamma * interaction_sum + epsilon_a * sigma_i - epsilon_r * (sigma_i**3)
    )
    return d_sigma


def simulate_swarmalators(
    N, J, alpha, beta, gamma, epsilon_a, epsilon_r, R, D, T=100, dt=0.1
):
    """
    Swarmalator

    Args:
        N (int): Number of swarmalators
        J (float): Coupling strength
        alpha (float): Constant for x dynamics
        beta (float): Constant for x dynamics
        gamma (float): Constant for sigma dynamics
        epsilon_a (float): Constant for sigma dynamics
        epsilon_r (float): Constant for sigma dynamics
        R (float): Radius of interaction
        D (int): Dimension of the system
        T (int): Number of time steps
        dt (float): Time step size

    Returns:
        results_xi (list): List of length T, each element is a tensor of shape (N, D)
        results_sigma_i (list): List of length T, each element is a tensor of shape (N, D)

    Example:
    import torch
    from swarmalator import Swarmulator


    # Initialize the Swarmulator
    N = 100  # Number of agents
    D = 100  # Dimensionality of agents
    swarm = Swarmulator(N=N, D=D, heads=5)

    # Run a simple forward pass
    swarm.simulation(num_steps=10)

    # Print the final positions and orientations of the swarm agents
    print("Final positions (xi) of the agents:")
    print(swarm.xi)
    print("\nFinal orientations (oi) of the agents:")
    print(swarm.oi)
    """
    xi = 2 * torch.rand(N, 3) - 1
    sigma_i = torch.nn.functional.normalize(torch.randn(N, D), dim=1)

    results_xi = []
    results_sigma_i = []

    for t in range(T):
        for i in range(N):
            dx = function_for_x(
                xi,
                sigma_i,
                N,
                J,
                alpha,
                beta,
                gamma,
                epsilon_a,
                epsilon_r,
                R,
                D,
            )
            d_sigma = function_for_sigma(
                xi,
                sigma_i,
                N,
                J,
                alpha,
                beta,
                gamma,
                epsilon_a,
                epsilon_r,
                R,
                D,
            )

            # RK4 for xi
            k1_x = dt * dx
            k2_x = dt * function_for_x(
                xi + 0.5 * k1_x,
                sigma_i,
                N,
                J,
                alpha,
                beta,
                gamma,
                epsilon_a,
                epsilon_r,
                R,
                D,
            )
            k3_x = dt * function_for_x(
                xi + 0.5 * k2_x,
                sigma_i,
                N,
                J,
                alpha,
                beta,
                gamma,
                epsilon_a,
                epsilon_r,
                R,
                D,
            )
            k4_x = dt * function_for_x(
                xi + k3_x,
                sigma_i,
                N,
                J,
                alpha,
                beta,
                gamma,
                epsilon_a,
                epsilon_r,
                R,
                D,
            )
            xi = xi + (1 / 6) * (k1_x + 2 * k2_x + 2 * k3_x + k4_x)

            # RK4 for sigma_i
            k1_sigma = dt * d_sigma
            k2_sigma = dt * function_for_sigma(
                xi,
                sigma_i + 0.5 * k1_sigma,
                N,
                J,
                alpha,
                beta,
                gamma,
                epsilon_a,
                epsilon_r,
                R,
                D,
            )
            k3_sigma = dt * function_for_sigma(
                xi,
                sigma_i + 0.5 * k2_sigma,
                N,
                J,
                alpha,
                beta,
                gamma,
                epsilon_a,
                epsilon_r,
                R,
                D,
            )
            k4_sigma = dt * function_for_sigma(
                xi,
                sigma_i + k3_sigma,
                N,
                J,
                alpha,
                beta,
                gamma,
                epsilon_a,
                epsilon_r,
                R,
                D,
            )
            sigma_i = sigma_i + (1 / 6) * (
                k1_sigma + 2 * k2_sigma + 2 * k3_sigma + k4_sigma
            )
            sigma_i = torch.nn.functional.normalize(sigma_i, dim=1)

        results_xi.append(xi.clone())
        results_sigma_i.append(sigma_i.clone())

    return results_xi, results_sigma_i



from dataclasses import dataclass
from datetime import datetime
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple

import matplotlib.pyplot as plt
import numpy as np
import torch
import torch.nn.functional as F
from loguru import logger
from sklearn.datasets import make_moons
from torch import Tensor, nn
from torch.distributions import Categorical


@dataclass
class FlowConfig:
    """Configuration for Flow neural network.

    Attributes:
        dim: Input/output dimension
        hidden_dim: Hidden layer dimension
        learning_rate: Learning rate for optimizer
        n_iterations: Number of training iterations
        batch_size: Batch size for training
        noise_level: Noise level for moon dataset
        device: torch device for computation
    """

    dim: int = 2
    hidden_dim: int = 64
    learning_rate: float = 1e-2
    n_iterations: int = 10000
    batch_size: int = 256
    noise_level: float = 0.05
    device: torch.device = torch.device(
        "cuda" if torch.cuda.is_available() else "cpu"
    )


class Flow(nn.Module):
    """Neural network for modeling continuous normalizing flows.

    This class implements a neural network that learns the velocity field
    of a continuous normalizing flow, transforming a simple distribution
    (e.g., Gaussian) into a more complex target distribution.
    """

    def __init__(self, config: FlowConfig):
        """Initialize the Flow model.

        Args:
            config: Configuration object containing model parameters
        """
        super().__init__()
        logger.info(f"Initializing Flow model with config: {config}")

        self.config = config
        self.net = nn.Sequential(
            nn.Linear(config.dim + 1, config.hidden_dim),
            nn.ELU(),
            nn.Linear(config.hidden_dim, config.hidden_dim),
            nn.ELU(),
            nn.Linear(config.hidden_dim, config.hidden_dim),
            nn.ELU(),
            nn.Linear(config.hidden_dim, config.dim),
        ).to(config.device)

        logger.debug(f"Model architecture: {self.net}")

    def forward(self, t: Tensor, x_t: Tensor) -> Tensor:
        """Forward pass computing the velocity field at time t and position x_t.

        Args:
            t: Time tensor of shape (batch_size, 1)
            x_t: Position tensor of shape (batch_size, dim)

        Returns:
            Velocity field tensor of shape (batch_size, dim)
        """
        input_tensor = torch.cat((t, x_t), -1)
        return self.net(input_tensor)

    def step(self, x_t: Tensor, t_start: Tensor, t_end: Tensor) -> Tensor:
        """Perform one step of numerical integration using the midpoint method.

        Args:
            x_t: Current position tensor of shape (batch_size, dim)
            t_start: Start time of the step
            t_end: End time of the step

        Returns:
            New position tensor after integration step
        """
        t_start = t_start.view(1, 1).expand(x_t.shape[0], 1)
        dt = t_end - t_start

        # Midpoint method
        k1 = self(t=t_start, x_t=x_t)
        x_mid = x_t + k1 * dt / 2
        t_mid = t_start + dt / 2
        k2 = self(t=t_mid, x_t=x_mid)

        return x_t + k2 * dt

    def train_model(self, save_dir: Optional[Path] = None) -> List[float]:
        """Train the flow model using the moons dataset.

        Args:
            save_dir: Optional directory to save model checkpoints

        Returns:
            List of loss values during training
        """
        logger.info("Starting model training")
        optimizer = torch.optim.Adam(
            self.parameters(), self.config.learning_rate
        )
        loss_fn = nn.MSELoss()
        losses = []

        for iteration in range(self.config.n_iterations):
            # Generate data
            x_1 = torch.tensor(
                make_moons(
                    self.config.batch_size,
                    noise=self.config.noise_level,
                )[0],
                dtype=torch.float32,
            ).to(self.config.device)
            x_0 = torch.randn_like(x_1).to(self.config.device)
            t = torch.rand(len(x_1), 1).to(self.config.device)

            # Compute target points
            x_t = (1 - t) * x_0 + t * x_1
            dx_t = x_1 - x_0

            # Training step
            optimizer.zero_grad()
            loss = loss_fn(self(t=t, x_t=x_t), dx_t)
            loss.backward()
            optimizer.step()
            losses.append(loss.item())

            if iteration % 1000 == 0:
                logger.info(f"Iteration {iteration}, Loss: {loss.item():.6f}")

            if save_dir and iteration % 1000 == 0:
                self.save_checkpoint(save_dir / f"checkpoint_{iteration}.pt")

        logger.info("Training completed")
        return losses

    def visualize_flow(
        self,
        n_points: int = 300,
        n_steps: int = 8,
        save_path: Optional[Path] = None,
    ) -> Tuple[plt.Figure, np.ndarray]:
        """Visualize the learned flow transformation.

        Args:
            n_points: Number of points to visualize
            n_steps: Number of integration steps
            save_path: Optional path to save the visualization

        Returns:
            Tuple of (figure, axes array)
        """
        logger.info("Generating flow visualization")
        x = torch.randn(n_points, 2).to(self.config.device)
        time_steps = torch.linspace(0, 1.0, n_steps + 1)

        fig, axes = plt.subplots(
            1, n_steps + 1, figsize=(30, 4), sharex=True, sharey=True
        )

        axes[0].scatter(x.cpu().detach()[:, 0], x.cpu().detach()[:, 1], s=10)
        axes[0].set_title(f"t = {time_steps[0]:.2f}")
        axes[0].set_xlim(-3.0, 3.0)
        axes[0].set_ylim(-3.0, 3.0)

        for i in range(n_steps):
            x = self.step(x_t=x, t_start=time_steps[i], t_end=time_steps[i + 1])
            axes[i + 1].scatter(
                x.cpu().detach()[:, 0], x.cpu().detach()[:, 1], s=10
            )
            axes[i + 1].set_title(f"t = {time_steps[i + 1]:.2f}")

        plt.tight_layout()

        if save_path:
            plt.savefig(save_path)
            logger.info(f"Visualization saved to {save_path}")

        return fig, axes

    def save_checkpoint(self, path: Path) -> None:
        """Save model checkpoint.

        Args:
            path: Path to save the checkpoint
        """
        path.parent.mkdir(parents=True, exist_ok=True)
        torch.save(
            {
                "model_state_dict": self.state_dict(),
                "config": self.config,
            },
            path,
        )
        logger.info(f"Checkpoint saved to {path}")

    @classmethod
    def load_checkpoint(cls, path: Path) -> "Flow":
        """Load model from checkpoint.

        Args:
            path: Path to the checkpoint file

        Returns:
            Loaded Flow model
        """
        checkpoint = torch.load(path)
        model = cls(checkpoint["config"])
        model.load_state_dict(checkpoint["model_state_dict"])
        logger.info(f"Model loaded from {path}")
        return model


# # Example usage
# if __name__ == "__main__":
#     # Setup logging
#     logger.add("flow_model.log", rotation="500 MB")

#     # Initialize and train model
#     config = FlowConfig()
#     flow = Flow(config)
#     losses = flow.train_model(save_dir=Path("checkpoints"))

#     # Visualize results
#     flow.visualize_flow(save_path=Path("flow_visualization.png"))


@dataclass
class MixtureFlowConfig:
    """Configuration for Mixture of Flows neural network.

    Attributes:
        n_experts: Number of flow experts in the mixture
        dim: Input/output dimension
        hidden_dim: Hidden layer dimension for each expert
        gating_hidden_dim: Hidden dimension for gating network
        learning_rate: Learning rate for optimizer
        n_iterations: Number of training iterations
        batch_size: Batch size for training
        noise_level: Noise level for moon dataset
        expert_dropout: Dropout rate for expert selection
        checkpoint_interval: Number of iterations between checkpoints
        device: torch device for computation
        logging_interval: Number of iterations between logging
    """

    n_experts: int = 4
    dim: int = 2
    hidden_dim: int = 64
    gating_hidden_dim: int = 32
    learning_rate: float = 1e-3
    n_iterations: int = 10000
    batch_size: int = 256
    noise_level: float = 0.05
    expert_dropout: float = 0.1
    checkpoint_interval: int = 1000
    logging_interval: int = 100
    device: torch.device = torch.device(
        "cuda" if torch.cuda.is_available() else "cpu"
    )


class GatingNetwork(nn.Module):
    """Neural network for selecting which flow expert to use.

    The gating network takes the current state and time as input and outputs
    a probability distribution over experts.
    """

    def __init__(self, config: MixtureFlowConfig):
        """Initialize the gating network.

        Args:
            config: Configuration object containing model parameters
        """
        super().__init__()
        self.config = config

        self.net = nn.Sequential(
            nn.Linear(config.dim + 1, config.gating_hidden_dim),
            nn.LayerNorm(config.gating_hidden_dim),
            nn.ReLU(),
            nn.Dropout(config.expert_dropout),
            nn.Linear(config.gating_hidden_dim, config.gating_hidden_dim),
            nn.LayerNorm(config.gating_hidden_dim),
            nn.ReLU(),
            nn.Linear(config.gating_hidden_dim, config.n_experts),
        ).to(config.device)

        logger.debug(f"Gating network architecture: {self.net}")

    def forward(self, t: Tensor, x_t: Tensor) -> Tensor:
        """Forward pass computing expert selection probabilities.

        Args:
            t: Time tensor of shape (batch_size, 1)
            x_t: Position tensor of shape (batch_size, dim)

        Returns:
            Expert selection probabilities tensor of shape (batch_size, n_experts)
        """
        input_tensor = torch.cat((t, x_t), -1)
        logits = self.net(input_tensor)
        return F.softmax(logits, dim=-1)


class ExpertFlow(nn.Module):
    """Individual flow expert network."""

    def __init__(self, config: MixtureFlowConfig, expert_id: int):
        """Initialize an expert flow network.

        Args:
            config: Configuration object containing model parameters
            expert_id: Unique identifier for this expert
        """
        super().__init__()
        self.config = config
        self.expert_id = expert_id

        self.net = nn.Sequential(
            nn.Linear(config.dim + 1, config.hidden_dim),
            nn.LayerNorm(config.hidden_dim),
            nn.ELU(),
            nn.Linear(config.hidden_dim, config.hidden_dim),
            nn.LayerNorm(config.hidden_dim),
            nn.ELU(),
            nn.Linear(config.hidden_dim, config.hidden_dim),
            nn.LayerNorm(config.hidden_dim),
            nn.ELU(),
            nn.Linear(config.hidden_dim, config.dim),
        ).to(config.device)

        logger.debug(f"Expert {expert_id} architecture: {self.net}")

    def forward(self, t: Tensor, x_t: Tensor) -> Tensor:
        """Forward pass computing the velocity field for this expert.

        Args:
            t: Time tensor of shape (batch_size, 1)
            x_t: Position tensor of shape (batch_size, dim)

        Returns:
            Velocity field tensor of shape (batch_size, dim)
        """
        input_tensor = torch.cat((t, x_t), -1)
        return self.net(input_tensor)


class MixtureFlow(nn.Module):
    """Mixture of Flows model combining multiple flow experts with a gating network."""

    def __init__(self, config: MixtureFlowConfig):
        """Initialize the Mixture of Flows model.

        Args:
            config: Configuration object containing model parameters
        """
        super().__init__()
        self.config = config
        logger.info(
            f"Initializing Mixture of Flows model with config: {config}"
        )

        # Initialize gating network
        self.gating = GatingNetwork(config)

        # Initialize flow experts
        self.experts = nn.ModuleList(
            [ExpertFlow(config, i) for i in range(config.n_experts)]
        )

        self._setup_logging()

    def _setup_logging(self) -> None:
        """Setup logging configuration."""
        timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")
        log_path = Path(f"logs/mixture_flow_{timestamp}.log")
        log_path.parent.mkdir(parents=True, exist_ok=True)
        logger.add(log_path, rotation="500 MB")

    def forward(self, t: Tensor, x_t: Tensor) -> Tuple[Tensor, Tensor]:
        """Forward pass computing the combined velocity field.

        Args:
            t: Time tensor of shape (batch_size, 1)
            x_t: Position tensor of shape (batch_size, dim)

        Returns:
            Tuple of (combined velocity field, expert weights) tensors
        """
        # Get expert selection probabilities
        expert_weights = self.gating(t, x_t)

        # Compute velocity field for each expert
        expert_outputs = torch.stack(
            [expert(t, x_t) for expert in self.experts], dim=1
        )  # Shape: (batch_size, n_experts, dim)

        # Combine expert outputs using weights
        combined_output = torch.sum(
            expert_outputs * expert_weights.unsqueeze(-1), dim=1
        )

        return combined_output, expert_weights

    def step(
        self, x_t: Tensor, t_start: Tensor, t_end: Tensor
    ) -> Tuple[Tensor, Tensor]:
        """Perform one step of numerical integration using the midpoint method.

        Args:
            x_t: Current position tensor of shape (batch_size, dim)
            t_start: Start time of the step
            t_end: End time of the step

        Returns:
            Tuple of (new position tensor, expert weights)
        """
        t_start = t_start.view(1, 1).expand(x_t.shape[0], 1)
        dt = t_end - t_start

        # Midpoint method
        k1, w1 = self(t=t_start, x_t=x_t)
        x_mid = x_t + k1 * dt / 2
        t_mid = t_start + dt / 2
        k2, w2 = self(t=t_mid, x_t=x_mid)

        # Average expert weights across the step
        avg_weights = (w1 + w2) / 2

        return x_t + k2 * dt, avg_weights

    def train_model(
        self, save_dir: Optional[Path] = None
    ) -> Dict[str, List[float]]:
        """Train the mixture of flows model.

        Args:
            save_dir: Optional directory to save model checkpoints

        Returns:
            Dictionary containing training metrics
        """
        logger.info("Starting model training")
        save_dir = save_dir or Path("checkpoints")
        save_dir.mkdir(parents=True, exist_ok=True)

        optimizer = torch.optim.Adam(
            self.parameters(), self.config.learning_rate
        )
        loss_fn = nn.MSELoss()

        metrics = {
            "total_loss": [],
            "mse_loss": [],
            "entropy_loss": [],
            "expert_usage": np.zeros(self.config.n_experts),
        }

        for iteration in range(self.config.n_iterations):
            # Generate data
            x_1 = torch.tensor(
                make_moons(
                    self.config.batch_size,
                    noise=self.config.noise_level,
                )[0],
                dtype=torch.float32,
            ).to(self.config.device)
            x_0 = torch.randn_like(x_1).to(self.config.device)
            t = torch.rand(len(x_1), 1).to(self.config.device)

            # Compute target points
            x_t = (1 - t) * x_0 + t * x_1
            dx_t = x_1 - x_0

            # Training step
            optimizer.zero_grad()
            pred_dx_t, expert_weights = self(t=t, x_t=x_t)

            # Compute losses
            mse_loss = loss_fn(pred_dx_t, dx_t)
            entropy_loss = Categorical(probs=expert_weights).entropy().mean()
            total_loss = (
                mse_loss - 0.1 * entropy_loss
            )  # Encourage expert specialization

            total_loss.backward()
            optimizer.step()

            # Update metrics
            metrics["total_loss"].append(total_loss.item())
            metrics["mse_loss"].append(mse_loss.item())
            metrics["entropy_loss"].append(entropy_loss.item())
            metrics["expert_usage"] += (
                expert_weights.mean(0).cpu().detach().numpy()
            )

            # Logging
            if iteration % self.config.logging_interval == 0:
                self._log_training_status(iteration, metrics)

            # Checkpointing
            if save_dir and iteration % self.config.checkpoint_interval == 0:
                self.save_checkpoint(save_dir / f"checkpoint_{iteration}.pt")

        logger.info("Training completed")
        self._log_final_metrics(metrics)
        return metrics

    def _log_training_status(
        self, iteration: int, metrics: Dict[str, Any]
    ) -> None:
        """Log training status.

        Args:
            iteration: Current training iteration
            metrics: Dictionary of training metrics
        """
        logger.info(
            f"Iteration {iteration}: "
            f"Total Loss: {metrics['total_loss'][-1]:.6f}, "
            f"MSE Loss: {metrics['mse_loss'][-1]:.6f}, "
            f"Entropy Loss: {metrics['entropy_loss'][-1]:.6f}"
        )

    def _log_final_metrics(self, metrics: Dict[str, Any]) -> None:
        """Log final training metrics.

        Args:
            metrics: Dictionary of training metrics
        """
        expert_usage = metrics["expert_usage"] / self.config.n_iterations
        logger.info("Final expert usage distribution:")
        for i, usage in enumerate(expert_usage):
            logger.info(f"Expert {i}: {usage:.3f}")

    def save_checkpoint(self, path: Path) -> None:
        """Save model checkpoint.

        Args:
            path: Path to save the checkpoint
        """
        checkpoint = {
            "model_state_dict": self.state_dict(),
            "config": self.config,
            "timestamp": datetime.now().isoformat(),
        }
        torch.save(checkpoint, path)
        logger.info(f"Checkpoint saved to {path}")

    @classmethod
    def load_checkpoint(cls, path: Path) -> "MixtureFlow":
        """Load model from checkpoint.

        Args:
            path: Path to the checkpoint file

        Returns:
            Loaded MixtureFlow model
        """
        checkpoint = torch.load(path)
        model = cls(checkpoint["config"])
        model.load_state_dict(checkpoint["model_state_dict"])
        logger.info(
            f"Model loaded from {path} (saved at {checkpoint['timestamp']})"
        )
        return model

    def visualize_flow(
        self,
        n_points: int = 300,
        n_steps: int = 8,
        save_path: Optional[Path] = None,
    ) -> Tuple[plt.Figure, np.ndarray]:
        """Visualize the learned flow transformation with expert contributions.

        Args:
            n_points: Number of points to visualize
            n_steps: Number of integration steps
            save_path: Optional path to save the visualization

        Returns:
            Tuple of (figure, axes array)
        """
        logger.info("Generating flow visualization")
        x = torch.randn(n_points, 2).to(self.config.device)
        time_steps = torch.linspace(0, 1.0, n_steps + 1)

        # Create subplot for each timestep plus expert usage plot
        fig, axes = plt.subplots(2, n_steps + 1, figsize=(30, 8))

        # Plot initial distribution
        axes[0, 0].scatter(x.cpu().detach()[:, 0], x.cpu().detach()[:, 1], s=10)
        axes[0, 0].set_title(f"t = {time_steps[0]:.2f}")
        axes[0, 0].set_xlim(-3.0, 3.0)
        axes[0, 0].set_ylim(-3.0, 3.0)

        # Storage for expert weights over time
        expert_weights_over_time = []

        for i in range(n_steps):
            x, weights = self.step(
                x_t=x, t_start=time_steps[i], t_end=time_steps[i + 1]
            )
            expert_weights_over_time.append(
                weights.mean(0).cpu().detach().numpy()
            )

            # Plot points
            axes[0, i + 1].scatter(
                x.cpu().detach()[:, 0], x.cpu().detach()[:, 1], s=10
            )
            axes[0, i + 1].set_title(f"t = {time_steps[i + 1]:.2f}")

            # Plot expert usage
            expert_weights = np.array(expert_weights_over_time)
            for e in range(self.config.n_experts):
                axes[1, i + 1].bar(e, expert_weights[-1, e])
            axes[1, i + 1].set_title("Expert Usage")
            axes[1, i + 1].set_ylim(0, 1)

        plt.tight_layout()

        if save_path:
            save_path.parent.mkdir(parents=True, exist_ok=True)
            plt.savefig(save_path)
            logger.info(f"Visualization saved to {save_path}")

        return fig, axes

    def visualize_expert_specialization(
        self, grid_size: int = 20, save_path: Optional[Path] = None
    ) -> plt.Figure:
        """Visualize how different experts specialize in different regions of the space.

        Args:
            grid_size: Number of points along each dimension in the visualization grid
            save_path: Optional path to save the visualization

        Returns:
            Matplotlib figure object
        """
        logger.info("Generating expert specialization visualization")

        # Create grid of points
        x = np.linspace(-3, 3, grid_size)
        y = np.linspace(-3, 3, grid_size)
        X, Y = np.meshgrid(x, y)
        grid_points = torch.tensor(
            np.stack([X.flatten(), Y.flatten()], axis=1),
            dtype=torch.float32,
        ).to(self.config.device)

        # Sample different time points
        time_points = torch.tensor([0.0, 0.33, 0.66, 1.0]).to(
            self.config.device
        )

        fig, axes = plt.subplots(
            len(time_points), 1, figsize=(10, 4 * len(time_points))
        )
        if len(time_points) == 1:
            axes = [axes]

        for t_idx, t in enumerate(time_points):
            t_expanded = t.view(1, 1).expand(len(grid_points), 1)
            expert_weights = self.gating(t_expanded, grid_points)

            # Get dominant expert at each point
            dominant_experts = torch.argmax(expert_weights, dim=1).cpu().numpy()

            # Plot
            scatter = axes[t_idx].scatter(
                grid_points[:, 0].cpu(),
                grid_points[:, 1].cpu(),
                c=dominant_experts,
                cmap="tab10",
                alpha=0.6,
            )
            axes[t_idx].set_title(f"t = {t.item():.2f}")
            axes[t_idx].set_xlim(-3, 3)
            axes[t_idx].set_ylim(-3, 3)

            # Add colorbar
            plt.colorbar(
                scatter,
                ax=axes[t_idx],
                label="Dominant Expert",
                ticks=range(self.config.n_experts),
            )

        plt.tight_layout()

        if save_path:
            save_path.parent.mkdir(parents=True, exist_ok=True)
            plt.savefig(save_path)
            logger.info(
                f"Expert specialization visualization saved to {save_path}"
            )

        return fig

    def get_expert_statistics(self) -> Dict[str, Any]:
        """Compute statistics about expert usage and specialization.

        Returns:
            Dictionary containing various expert statistics
        """
        logger.info("Computing expert statistics")

        # Generate grid of points for analysis
        x = torch.linspace(-3, 3, 20)
        y = torch.linspace(-3, 3, 20)
        X, Y = torch.meshgrid(x, y, indexing="ij")
        grid_points = torch.stack([X.flatten(), Y.flatten()], dim=1).to(
            self.config.device
        )

        # Sample different time points
        time_points = torch.linspace(0, 1, 10).to(self.config.device)

        expert_stats = {
            "total_usage": torch.zeros(self.config.n_experts).to(
                self.config.device
            ),
            "max_confidence": torch.zeros(self.config.n_experts).to(
                self.config.device
            ),
            "avg_confidence": torch.zeros(self.config.n_experts).to(
                self.config.device
            ),
            "territory_size": torch.zeros(self.config.n_experts).to(
                self.config.device
            ),
        }

        total_points = len(grid_points) * len(time_points)

        for t in time_points:
            t_expanded = t.view(1, 1).expand(len(grid_points), 1)
            expert_weights = self.gating(t_expanded, grid_points)

            # Update statistics
            expert_stats["total_usage"] += expert_weights.sum(dim=0)
            expert_stats["max_confidence"] = torch.maximum(
                expert_stats["max_confidence"],
                expert_weights.max(dim=0)[0],
            )
            expert_stats["avg_confidence"] += expert_weights.sum(dim=0)
            expert_stats["territory_size"] += (
                (expert_weights == expert_weights.max(dim=1, keepdim=True)[0])
                .float()
                .sum(dim=0)
            )

        # Normalize statistics
        expert_stats["total_usage"] /= total_points
        expert_stats["avg_confidence"] /= total_points
        expert_stats["territory_size"] /= total_points

        # Convert to numpy for easier handling
        return {k: v.cpu() for k, v in expert_stats.items()}


# # Example usage
# if __name__ == "__main__":
#     # Setup configuration
#     config = MixtureFlowConfig()

#     # Initialize and train model
#     model = MixtureFlow(config)
#     training_metrics = model.train_model(save_dir=Path("checkpoints"))

#     # Visualize results
#     model.visualize_flow(
#         save_path=Path("visualizations/flow_evolution.png")
#     )
#     model.visualize_expert_specialization(
#         save_path=Path("visualizations/expert_specialization.png")
#     )

#     # Get and log expert statistics
#     expert_stats = model.get_expert_statistics()
#     logger.info("\nExpert Statistics:")
#     for stat_name, values in expert_stats.items():
#         logger.info(f"\n{stat_name}:")
#         for expert_id, value in enumerate(values):
#             logger.info(f"Expert {expert_id}: {value:.3f}")



import torch
import torch.nn as nn
import torch.nn.functional as F


class CustomMLP(nn.Module):
    """
    A customizable Multi-Layer Perceptron (MLP).

    Attributes:
        layers (nn.ModuleList): List of linear layers.
        activation_fn (callable): Activation function to be applied after each layer.
        dropout (float): Dropout probability for regularization.

    Parameters:
        layer_sizes (list of int): List of layer sizes including input and output layer.
        activation (str, optional): Type of activation function. Default is 'relu'.
        dropout (float, optional): Dropout probability. Default is 0.0 (no dropout).
    """

    def __init__(self, layer_sizes, activation="relu", dropout=0.0):
        super().__init__()

        # Validate input parameters
        if not isinstance(layer_sizes, list) or len(layer_sizes) < 2:
            raise ValueError(
                "layer_sizes must be a list with at least two integers"
                " representing input and output sizes."
            )
        if not all(isinstance(size, int) and size > 0 for size in layer_sizes):
            raise ValueError(
                "All elements in layer_sizes must be positive integers."
            )

        if dropout < 0.0 or dropout > 1.0:
            raise ValueError("dropout must be a float between 0.0 and 1.0")

        # Define the activation function
        if activation == "relu":
            self.activation_fn = F.relu
        elif activation == "sigmoid":
            self.activation_fn = torch.sigmoid
        elif activation == "tanh":
            self.activation_fn = torch.tanh
        else:
            raise ValueError(
                "Unsupported activation function. Supported: 'relu', 'sigmoid',"
                " 'tanh'."
            )

        # Create layers
        self.layers = nn.ModuleList()
        for i in range(len(layer_sizes) - 1):
            self.layers.append(nn.Linear(layer_sizes[i], layer_sizes[i + 1]))

        # Dropout layer
        self.dropout = nn.Dropout(dropout)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Forward pass through the MLP.

        Parameters:
            x (torch.Tensor): Input tensor.

        Returns:
            torch.Tensor: Output tensor.
        """
        for i in range(len(self.layers) - 1):
            x = self.layers[i](x)
            x = self.activation_fn(x)
            x = self.dropout(x)
        x = self.layers[-1](x)  # No activation or dropout on the last layer
        return x


# Example Usage:
# mlp = CustomMLP(layer_sizes=[10, 5, 2], activation='relu', dropout=0.5)
# input_data = torch.randn(1, 10)
# output = mlp(input_data)



from torch import Tensor, nn


class PaloLDP(nn.Module):
    """
    Implementation of the PaloLDP module.

    Args:
        dim (int): The dimension of the input tensor.
        channels (int, optional): The number of input channels. Defaults to 1.
    """

    def __init__(
        self,
        dim: int,
        channels: int = 1,
    ):
        super().__init__()
        self.dim = dim
        self.channels = channels

        self.pointwise_conv = nn.Conv2d(
            in_channels=channels,
            out_channels=channels,
            kernel_size=1,
            stride=1,
            padding=0,
        )

        self.gelu = nn.GELU()

        # Depthwise convolution
        self.depthwise_conv = nn.Conv2d(
            in_channels=channels,
            out_channels=channels,
            kernel_size=3,
            stride=1,
            padding=1,
            groups=channels,
        )

        # LayerNorm
        self.norm = nn.LayerNorm(dim)

        # Depthwise convolution with stride = 2
        self.depthwise_conv_stride = nn.Conv2d(
            in_channels=channels,
            out_channels=channels,
            kernel_size=3,
            stride=2,
            padding=1,
            groups=channels,
        )

    def forward(self, x: Tensor) -> Tensor:
        """
        Forward pass of the PaloLDP module.

        Args:
            x (Tensor): The input tensor of shape (B, C, H, W).

        Returns:
            Tensor: The output tensor of shape (B, C, H', W').
        """
        b, c, h, w = x.shape

        x = self.pointwise_conv(x)
        print(x.shape)  # torch.Size([2, 1, 4, 4]

        x = self.gelu(x)
        print(x.shape)  # torch.Size([2, 1, 4, 4]

        x = self.pointwise_conv(x)
        print(x.shape)  # torch.Size([2, 1, 4, 4]

        # Depthwise convolution with 1 stide
        x = self.depthwise_conv(x)
        print(x.shape)

        # Norm
        x = self.norm(x)
        print(x.shape)

        # Pointwise convolution
        x = self.pointwise_conv(x)
        print(x.shape)

        # Norm
        x = self.norm(x)  # + skip
        print(x.shape)

        # Depthwise convolution with 2 stide
        x = self.depthwise_conv_stride(x)
        print(x.shape)

        # Norm
        b, c, h, w = x.shape
        # x = self.norm(x)
        x = nn.LayerNorm(w)(x)

        # Pointwise convolution
        x = self.pointwise_conv(x)

        # Norm
        b, c, h, w = x.shape
        x = nn.LayerNorm(w)(x)

        return x



from copy import deepcopy

import torch
import torch.nn.functional as F
from einops import rearrange
from torch import Tensor, nn


def freeze_all_layers(module):
    for param in module.parameters():
        param.reqires_grad = False


def log(t, eps=1e-20):
    return torch.log(t.clamp(min=eps))


def log_prob(prob, indices, eps=1e-20):
    indices = rearrange(indices, "... -> ... 1")
    log_probs = log(prob.gather(-1, indices), eps=eps)
    return rearrange(log_probs, "... 1 -> ...")


def log_prob_from_model_and_seq(model, seq):
    logits = model(seq)
    prob = logits.softmax(dim=-1)
    return log_prob(prob, seq)


class DPO(nn.Module):
    """
    Deep Policy Optimization (DPO) module.

    Args:
        model (nn.Module): The policy model.
        beta (float, optional): The beta parameter. Defaults to 0.1.
    """

    def __init__(self, model: nn.Module, *, beta: float = 0.1):
        super().__init__()
        self.policy_model = model

        self.ref_model = deepcopy(model)
        freeze_all_layers(self.ref_model)

        self.beta = beta

    def parameters(self):
        return self.policy_model.parameters()

    def forward(self, preferred_seq: Tensor, unpreferred_seq: Tensor):
        """
        Forward pass of the DPO module.

        Args:
            preferred_seq (torch.Tensor): The preferred sequence.
            unpreferred_seq (torch.Tensor): The unpreferred sequence.

        Returns:
            torch.Tensor: The loss value.
        """
        assert preferred_seq.ndim == 2
        assert preferred_seq.shape == unpreferred_seq.shape

        """
        Following Appendix B in https://arxiv.org/abs/2305.18290
        """

        with torch.no_grad():
            self.ref_model.eval()
            ref_preferred_logprob = log_prob_from_model_and_seq(
                self.ref_model, preferred_seq
            )
            ref_unpreferred_logprob = log_prob_from_model_and_seq(
                self.ref_model, unpreferred_seq
            )

        policy_preferred_logprob = log_prob_from_model_and_seq(
            self.policy_model, preferred_seq
        )
        policy_unpreferred_logprob = log_prob_from_model_and_seq(
            self.policy_model, unpreferred_seq
        )

        policy_logratios = policy_preferred_logprob - policy_unpreferred_logprob
        ref_logratios = ref_preferred_logprob - ref_unpreferred_logprob

        losses = -F.logsigmoid(self.beta * (policy_logratios - ref_logratios))

        return losses.mean()



import torch
from torch import nn


class TripleSkipBlock(nn.Module):
    def __init__(self, submodule1, submodule2, submodule3):
        """
        TripleSkipBlock class represents a block that performs triple skip connections.

        Args:
            submodule1 (nn.Module): The first submodule.
            submodule2 (nn.Module): The second submodule.
            submodule3 (nn.Module): The third submodule.
        """
        super().__init__()
        self.submodule1 = submodule1
        self.submodule2 = submodule2
        self.submodule3 = submodule3

    def forward(self, x: torch.Tensor):
        """
        Forward pass of the TripleSkipBlock.

        Args:
            x (torch.Tensor): The input tensor.

        Returns:
            torch.Tensor: The output tensor after applying triple skip connections.
        """
        return x + self.submodule1(x + self.submodule2(x + self.submodule(x)))



from torch import nn, Tensor


class FractoralNorm(nn.Module):
    """
    FractoralNorm module applies LayerNorm to the input tensor multiple times in a row.

    Args:
        dim (int): Number of features in the input tensor.
        depth (int): Number of times to apply LayerNorm.
    """

    def __init__(self, dim: int, depth: int, *args, **kwargs):
        super().__init__()

        self.layers = nn.ModuleList(
            [nn.LayerNorm(dim, *args, **kwargs) for _ in range(depth)]
        )

    def forward(self, x: Tensor) -> Tensor:
        """
        Forward pass of the FractoralNorm module.

        Args:
            x (Tensor): Input tensor.

        Returns:
            Tensor: Output tensor after applying LayerNorm multiple times.
        """
        for layer in self.layers:
            x = layer(x)
        return x



from torch import nn, Tensor


class PEG(nn.Module):
    """
    PEG (Positional Encoding Generator) module.

    Args:
        dim (int): The input dimension.
        kernel_size (int, optional): The size of the convolutional kernel. Defaults to 3.
    """

    def __init__(self, dim: int, kernel_size: int = 3):
        super().__init__()
        self.proj = nn.Conv2d(
            dim,
            dim,
            kernel_size=kernel_size,
            padding=kernel_size // 2,
            groups=dim,
            stride=1,
        )

    def forward(self, x: Tensor):
        """
        Forward pass of the PEG module.

        Args:
            x (Tensor): The input tensor.

        Returns:
            Tensor: The output tensor.
        """
        return self.proj(x) + x
